#----------------------------------------------------------------------------
# Name:         ldmStorageFolder.py
# Purpose:      storage folder content
#
# Author:       Walter Obweger
#
# Created:      20200322
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import logging
import traceback

import os
from datetime import datetime as dt

from optparse import OptionParser

from lindworm import __version__
from lindworm.ldmStorage import ldmStorage
from lindworm.ldmOS import getSha

class ldmStorageFolder(ldmStorage):
    def __init__(self,iModeRev=0,sLogger='srgFld',iLv=0,iVerbose=0):
        """constructor
        ### parameter
            iModeRev    ... reverse processing
                    1   ... FILO, first in last out
                    0   ... FIFO, first in first out
            sLogger     ... log origin
            iLv         ... logging level
            iVerbose    ... higher values add more logs
        """
        ldmStorage.__init__(self,iModeRev=iModeRev,
                            sLogger=sLogger,iLv=iLv,
                            iVerbose=iVerbose)
    def __initCfg__(self):
        """initialize configuration properties
        """
        try:
            ldmStorage.__initCfg__(self)
            # +++++ beg:initialize
            sOrg='ldmStorage::__initCfg__'
            self.logDbg('beg:%s',sOrg)
            # ----- end:initialize
            # +++++ beg:default configuration
            self.dCfgDft['lSkipDN']=['CVS','.git']
            self.dCfgDft['sTimePrecision']='seconds'
            self.dCfgDft['sStatNew']='zNew'
            self.dCfgDft['sStatUse']='zUse'
            self.dCfgDft['sStatMod']='zMod'
            #self.dCfgDft['zPrecision']='minutes'
            # ----- end:default configuration
            self.logDbg('end:%s',sOrg)
            return 1
        except:
            self.logTB()
            return -1
    def __initPrc__(self):
        """initialize processing properties
        """
        try:
            ldmStorage.__initPrc__(self)
            # +++++ beg:initialize
            sOrg='ldmStorageFolder::__initDat__'
            self.logDbg('beg:%s',sOrg)
            # ----- end:initialize
            # +++++ beg:initialize data
            self.iCnt=0             # count of directories
            self.iAct=0             # 
            self.iSrcLen=0          # source DN length, used to build relative
            self.lCnt=[]            # files in relative DN
            self.lFdr=[]            # folder list as they arrived, relative DN
            self.dFolder={}         # content cache, relative DN
            # ----- end:initialize data
            self.logDbg('end:%s',sOrg)
            return 1
        except:
            self.logTB()
            return -1
    def clrEnd(self):
        """clear stored file names to process
        """
        iRet=ldmStorage.clrEnd(self)
        self.dCntEnd=None
        return iRet
    def getDatRel(self,sRelDN,sBaseDN=None):
        """get relative content from data
        ### parameter
            sRelDN  ... relative DN
            sBaseDN ... base DN
        ### return
            iRet
                >0  ... okay processing done
                =0  ... okay nop
                <0  ... error
            dDN ... directory content
                dict  ... {'.':...}
                None  ... invalid content
        """
        try:
            # +++++ beg:get relative path list
            iR,dDN=self.getDat(sDef=sBaseDN)
            if iR<1:
                self.oLog.error('getDatRel: sFN:%s iR:%d dDat:%r'%(sBaseDN,iR,self.dDat))
                return -1,None
            if sRelDN=='./':
                lRelDN=['.']
            else:
                lRelDN=sRelDN.split('/')
            #if self.iVerbose>0:
            #    self.logDbg('            lRelDN:%r',lRelDN)
            # ----- end:get relative path list
            # +++++ beg:get dDN storage
            for sDN in lRelDN[1:]:
                if sDN not in dDN:
                    return 0,None
                dDN=dDN[sDN]
            return 1,dDN
            # ----- end:get dDN storage
        except:
            self.logTB()
            return -1,None
    def getDatContent(self,sRoot):
        """get content from data
        ### parameter
            sRoot   ... directory
        ### return
            sCurDN  ... current directory name, relative
            sDN     ... directory content
            sFN     ... file content
        """
        try:
            if self.iVerbose>90:
                self.logDbg('        sRoot:%s',sRoot)
            # +++++ beg:calc relative DN posix style
            sCurDN=sRoot[self.iSrcLen:].replace(os.sep,'/')
            if sCurDN=='':
                sCurDN='./'
            elif sCurDN[0]=='/':
                sCurDN='.'+sCurDN
            else:
                sCurDN='./'+sCurDN
            if self.iVerbose>0:
                self.logDbg('        sCurDN:%s',sCurDN)
            # ----- end:calc relative DN posix style
            # +++++ beg:prepare dDN storage
            iR,dDN=self.getDat()
            if iR<1:
                self.oLog.error('getDat() iR:%d dDN:%r'%(iR,dDN))
                dDN={}
            else:
                if self.iVerbose>0:
                    lKey=list(dDN.keys())
                    lKey.sort()
                    self.logDbg('            getDat() iR:%d dDN keys:%r',iR,lKey)
                if self.iVerbose>95:
                    self.logDbg('            getDat() iR:%d dDN:%r',iR,dDN)
            if sCurDN=='./':
                lCurDN=['.']
                if '.' not in dDN:
                    dDN['.']={}
            else:
                lCurDN=sCurDN.split('/')
            if self.iVerbose>0:
                self.logDbg('            lCurDN:%r',lCurDN)
            # ----- end:prepare dDN storage
            # +++++ beg:get dDN storage
            for sDN in lCurDN[1:]:
                if sDN not in dDN:
                    dDN[sDN]={'.':{}}
                dDN=dDN[sDN]
            dFN=dDN['.']
            # ----- end:get dDN storage
            # +++++ beg:build cache
            self.iCnt+=1
            self.lFdr.append(sCurDN)
            self.dFolder[sCurDN]=dDN
            # ----- end:build cache
            return sCurDN,dDN,dFN
        except:
            self.logTB()
            return None,None,None
    def prcBeg(self,sDef,oRef=None,**kwargs):
        """processing begin, add file name and objects to stack
        ### parameter
            sSrcDN  ... source directory name : str
            oDat    ... data object
            oRef    ... reference object
            kwargs  ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmStorageFolder::prcBeg'
            # ----- end:initialize
            # +++++ beg:processing begin
            self.logDbg('beg:%s sDef:%s',sOrg,sDef)
            if self.iVerbose>0:
                self.logDbg('    oRef:%r kwargs:%r',oRef,kwargs)
            dRef={
                'iAct':self.iAct,
                'oRef':oRef,
            }
            iRet=ldmStorage.prcBeg(self,sDef,oRef=dRef,**kwargs)
            if iRet>0:
                self.iAct=0
            self.logDbg('end:%s iRet:%d',sOrg,iRet)
            # ----- end:processing begin
            return iRet
        except:
            self.logTB()
            return -1
    def prcExc(self,oNty=None,**kwargs):
        """processing execution
        ### parameter
            **kwargs ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmStorageFolder::prcExc'
            # ----- end:initialize
            # +++++ beg:get definition
            iRet,sDef=self.getDef()
            if iRet<1:
                self.logDbg('skp:%s iRet:%d empty stack',sOrg,iRet)
                return iRet
            self.logDbg('beg:%s sDef:%s',sOrg,sDef)
            if self.iVerbose>0:
                self.logDbg('    kwargs:%r',kwargs)
            self.iSrcLen=len(sDef)
            # ----- end:get definition
            # +++++ beg:get folder list to skip
            iR,lSkipDN=self.getCfg('lSkipDN')
            if self.iVerbose>0:
                self.logDbg('    iR:%d lSkipDN:%r',iR,lSkipDN)
            # ----- end:get folder list to skip
            # +++++ beg:get configuration
            iR,sTimePrecision=self.getCfg('sTimePrecision')
            iR,sStatNew=self.getCfg('sStatNew',oDft=0)
            iR,sStatMod=self.getCfg('sStatMod',oDft=0)
            iR,sStatUse=self.getCfg('sStatUse',oDft=0)
            # ----- end:get configuration
            # +++++ beg:walk through directory recursive
            for sRoot,lDN,lFN in os.walk(sDef):
                if self.iVerbose>90:
                    self.logDbg('        sRoot:%s',sRoot)
                sCurDN,dDN,dFN=self.getDatContent(sRoot)
                if oNty is not None:
                    oNty.SetStatus(sCurDN)
                if self.iVerbose>0:
                    self.logDbg('        sCurDN:%s',sCurDN)
                #dFiles={}
                #dContent={'.':dFiles}
                #self.lFdr.append(sCurDN)
                #self.dFolder[sCurDN]=dContent
                if dFN is not None:
                    for sFN in lFN:
                        iR,oRef=self.getRef()
                        iR,dStat=self.getStat(sRoot,sFN,oRef,
                                        sTimePrecision=sTimePrecision,
                                        sStatNew=sStatNew,
                                        sStatMod=sStatMod,
                                        sStatUse=sStatUse,
                                        **kwargs)
                        if iR>0:
                            if self.iVerbose>0:
                                self.logDbg('          sFN:%s dStat:%r',sFN,dStat)
                            dFN[sFN]=dStat
                if self.iVerbose>5:
                    iLenFN=len(dFN)
                    self.logDbg('        sCurDN:%s iLenFN:%r',sCurDN,iLenFN)
                for sSkipDN in lSkipDN:
                    if sSkipDN in lDN:
                        lDN.remove(sSkipDN)
            # ----- end:walk through directory recursive
            # +++++ beg:finalize
            if self.iVerbose>90:
                self.logDbg('   :%s sDef:%s dDat:%r',sOrg,sDef,self.dDat)
            self.logDbg('end:%s iRet:%d',sOrg,iRet)
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def prcEnd(self,**kwargs):
        """processing end
        ### parameter
            *kwargs ... flexible keyword argument
            
            use updated properties to access current properties to be finalized
            self.sDefEnd    ... definition
            self.oArgEnd    ... keyword arguments passed at prcBeg
            self.oDatEnd    ... data object
            self.oRefEnd    ... reference object
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmStorageFolder::prcEnd'
            # ----- end:initialize
            # +++++ beg:processing end
            self.logDbg('beg:%s len(lFdr):%d kwargs:%r',sOrg,
                                len(self.lFdr),kwargs)
            iRet=ldmStorage.prcEnd(self,**kwargs)
            if iRet>0:
                # +++++ beg:revert current file info
                #self.iAct=iMarkerAct    # revert current file line back
                # ----- end:revert current file info
                if self.iVerbose>5:
                    self.logDbg('    %s iAct:%5d iCnt:%6d sDefEnd:%s',sOrg,
                                    self.iAct,self.iCnt,self.sDefEnd)
                if self.iVerbose>95:
                    self.logDbg('   :%s oDatEnd:%r',sOrg,
                                    self.oDatEnd)
                # +++++ beg:loop through folders
                iOfsFdr=0
                lCntKey=['sz','iDN','iFN',]
                dCntDef={
                    'sz':0,
                    'iDN':0,
                    'iFN':0,
                }
                for sRelDN in self.lFdr:
                    iR,dDN=self.getDatRel(sRelDN,sBaseDN=self.sDefEnd)
                    if self.iVerbose>5:
                        self.logDbg('   :%s sRelDN:%s iR:%d sDefEnd:%r',sOrg,
                                        sRelDN,iR,self.sDefEnd)
                    iSz=0
                    iLenFN=0
                    iLenDN=0
                    if iR>0:
                        if self.iVerbose>5:
                            self.logDbg('   :%s sRelDN:%s iR:%d len(dDN):%d',sOrg,
                                            sRelDN,iR,len(dDN))
                        if self.iVerbose>95:
                            self.logDbg('   :%s sRelDN:%s iR:%d dDN:%r',sOrg,
                                            sRelDN,iR,dDN)
                        dFN=dDN.get('.',None)
                        iLenDN=len(dDN)-1
                        if dFN is not None:
                            for sFN,dStat in dFN.items():
                                iSz+=dStat.get('sz',0)
                                iLenFN+=1
                    dCnt={
                        'sRelDN':sRelDN,
                        'sz':iSz,
                        'iDN':iLenDN,
                        'iFN':iLenFN,
                        }
                    self.lCnt.append(dCnt)
                    # +++++ beg:update count total
                    for sK in lCntKey:
                        dCntDef[sK]=dCntDef[sK]+dCnt[sK]
                    # ----- end:update count total
                    iOfsFdr+=1
                # ----- end:loop through folders
                # +++++ beg:post processing
                self.dCntEnd=dCntDef
                iR=self.prcPost(**kwargs)
                # ----- end:post processing
                if self.iVerbose>5:
                    self.logDbg('   :%s lCnt:%r',sOrg,self.lCnt)
            self.logDbg('end:%s iRet:%d',sOrg,iRet)
            # ----- end:processing end
            return iRet
        except:
            self.logTB()
            return -1
    def prcPost(self,**kwargs):
        """processing end
            **kwargs ... flexible keyword argument
                iShaMB  ... data to calculate sha in MB 
        ### return
            iRet    ... return code
                >0  ... okay processing done
                =0  ... okay nop
                <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmStorageFolder::prcPost'
            iShaMB=kwargs.get('iShaMB',0)
            # ----- end:initialize
            if self.sDefEnd is None:
                self.logDbg('skp:%s iRet:%d empty definition,'
                                'prcEnd has to be call before',sOrg,iRet)
                return iRet
            # +++++ beg:folder post processing
            iOfsFld=-1
            iCntFN=0
            self.logDbg('beg:%s sDefEnd:%s len(lFdr):%d kwargs:%r',sOrg,
                                self.sDefEnd,len(self.lFdr),kwargs)
            oNty=kwargs.get('oNty',None)
            if oNty is not None:
                self.logDbg('    dCntEnd:%r',self.dCntEnd)
                oNty.SetMax(self.dCntEnd['iFN'])
            for sRelDN in self.lFdr:
                iOfsFld+=1
                if sRelDN in self.dFolder:
                    if oNty is not None:
                        oNty.SetStatus(sRelDN)
                        oNty.IncStatus()
                    sDN='/'.join([self.sDefEnd,sRelDN])
                    dDN=self.dFolder[sRelDN]
                    if self.iVerbose>5:
                        self.logDbg('   :%s sRelDN:%s len(dDN):%d',sOrg,
                                        sRelDN,len(dDN))
                    if self.iVerbose>9:
                        self.logDbg('    dDN:%r',dDN)
                    if '.' in dDN:
                        dFN=dDN['.']
                        # +++++ beg:calc sha fingerprint
                        for sFN,dStat in dFN.items():
                            iCntFN+=1
                            if oNty is not None:
                                #oNty.SetStatus(sFN)
                                #oNty.IncStatus()
                                oNty.SetVal(iCntFN)
                            sSha=getSha(sFN,iMB=iShaMB,
                                        sDN=sDN,
                                        oLog=self.oLog)
                            if self.iVerbose>5:
                                self.logDbg('      sFN:%s sha:%s',sFN,
                                                sSha)
                            dStat['sha']=sSha
                        # ----- end:calc sha fingerprint
            # ----- end:folder post processing
            #if oNty is not None:
            #S    oNty.clrStatus()
            self.logDbg('end:%s iRet:%d',sOrg,iRet)
            return iRet
        except:
            self.logTB()
            return -1
    def getStatDct(self,sFN,oStat,dStat,oRef,**kwargs):
        return 0
    def getStatTimeStamp(self,sFN,oStat,dStat,oRef,
                    sTimePrecision='minutes',
                    sStatNew=0,
                    sStatMod=0,
                    sStatUse=0,
                    **kwargs):
        try:
            iRet=0
            if sStatNew:
                oTmp=dt.fromtimestamp(oStat.st_ctime)
                dStat[sStatNew]=oTmp.isoformat(timespec=sTimePrecision)
                iRet+=1
            if sStatMod:
                oTmp=dt.fromtimestamp(oStat.st_mtime)
                dStat[sStatMod]=oTmp.isoformat(timespec=sTimePrecision)
            if sStatUse:
                oAccess=dt.fromtimestamp(oStat.st_atime)
                dStat[sStatUse]=oTmp.isoformat(timespec=sTimePrecision)
            return iRet
        except:
            self.logTB()
        return 0
    def getStat(self,sDN,sFN,oRef,**kwargs):
        try:
            if sDN is not None:
                sTmpFN=os.path.join(sDN,sFN)
            else:
                sTmpFN=sFN
            if os.path.exists(sTmpFN):
                # +++++ beg:read file status information
                oStat=os.stat(sTmpFN)
                # ----- end:read file status information
                # +++++ beg:
                dStat={
                    'sz':oStat.st_size,
                }
                iR=self.getStatTimeStamp(sTmpFN,oStat,dStat,oRef,**kwargs)
                iR=self.getStatDct(sTmpFN,oStat,dStat,oRef,**kwargs)
                return 1,dStat
            return 0,None
        except:
            self.logTB()
            return -1,None

def execMain(sSrcDN='.',
                sBldDN=None,
                sBldFN=None,
                sCfgFN=None,
                iShaMB=1,
                iVerbose=0):
    """main function to perform conversion
    """
    try:
        # +++++ beg:
        # ----- end:
        # +++++ beg:
        logging.debug('beg:execMain')
        iRet=0
        # ----- end:
        # +++++ beg:
        oFld=ldmStorageFolder(iVerbose=iVerbose)
        iRet=oFld.loadCfg(sCfgFN)
        # ----- end:
        # +++++ beg:
        oFld.prcBeg(sSrcDN,oRef=None)
        oFld.prcExc()#(oGtrMD=self)
        iRetFdr=oFld.prcEnd(iShaMB=iShaMB)
        if iRetFdr>0:
            iRet+=1
        # ----- end:
        # +++++ beg:
        if iRetFdr>0:
            iRetSave=oFld.saveDat(sBldFN,sDN=sBldDN,
                                lKey=[oFld.sDefEnd],
                                sAtr='dDat')
        # ----- end:
        # +++++ beg:
        logging.debug('end:execMain iRet:%d'%(iRet))
        # ----- end:
        return iRet
    except:
        logging.error(traceback.format_exc())
        return -1

def main(args=None):
    # +++++ beg:
    # ----- end:
    
    # +++++ beg:init
    iVerbose=5                                          # 20190624 wro:set default verbose level
    # ----- end:init
    # +++++ beg:define CLI arguments
    usage = "usage: %prog [options]"
    oParser=OptionParser(usage,version="%prog "+__version__)
    oParser.add_option('-c','--cfgFN',dest='sCfgFN',
            default='ldmStorageFolderCfg.json',
            help='configuration file',metavar='pyGatherMDCfg.json',
            )
    oParser.add_option('','--srcDN',dest='sSrcDN',
            default='.',
            help='source folder',metavar='path/to/folder/to/read',
            )
    oParser.add_option('','--bldDN',dest='sBldDN',
            default=None,
            help='build directory',metavar='path/to/output/folder',
            )
    oParser.add_option('','--otBldFN',dest='sBldFN',
            default=None,
            help='build file',metavar='sng.json',
            )
    oParser.add_option('-l','--log',dest='sLogFN',
            default='./log/ldmStorageFolder.log',
            help='log filename',metavar='./log/ldmStorageFolder.log')
    oParser.add_option("-v", action="store_true", dest="verbose", default=True)
    oParser.add_option("-q", action="store_false", dest="verbose", default=True)
    # ----- end:define CLI arguments
    # +++++ beg:parse command line
    (opt,args)=oParser.parse_args(args=args)
    if opt.verbose:
        print("config FN:     %r"%(opt.sCfgFN))
        print("source DN:     %r"%(opt.sSrcDN))
        print(" build DN:     %r"%(opt.sBldDN))
        print(" build FN:     %r"%(opt.sBldFN))
        iVerbose=20
    # ----- end:parse command line
    # +++++ beg:prepare logging
    import lindworm.logUtil as logUtil
    logUtil.logInit(opt.sLogFN,iLevel=logging.DEBUG)
    # ----- end:prepare logging
    # +++++ beg:status
    print('logging file:%r'%(opt.sLogFN))
    # ----- end:status
    # +++++ beg:perform main action
    iRet=execMain(sSrcDN=opt.sSrcDN,
                sBldDN=opt.sBldDN,
                sBldFN=opt.sBldFN,
                sCfgFN=opt.sCfgFN,
                iVerbose=iVerbose)
    # ----- end:perform main action
    # +++++ beg:finished
    print('\nend processing')
    print('   iRet:%r'%(iRet))
    # ----- end:finished

if __name__=='__main__':
    # +++++ beg:call entry point
    main(args=None)
    # ----- end:call entry point



#----------------------------------------------------------------------------
# Name:         ldmArg.py
# Purpose:      argument class
#
# Author:       Walter Obweger
#
# Created:      20200412
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import sys
import traceback

class ldmArg:
    """argument container class to simplify command line argument setup 
    and allow GUI enhancement.
    support classical command line parser package OptionParser for py version < 2.7
    and replacement ArgumentParser.
    """
    MAP_TYPE={'i':int,'r':float,'s':str}
    def __init__(self,sUsage=None,sVer='',iVerbose=0):
        """constructor of argument container

        ### parameter
            sUsage  ... help string
            sVer    ... version number
        """
        self.lArg=[]
        self.dArg={}
        self.oOptRes=None
        self.iVerbose=iVerbose
        self.iIsDbg=1
        if sys.version_info  >= (2,7,0):
            from argparse import ArgumentParser
            self.oOpt=ArgumentParser()
            self.iMode=1
        else:
            from optparse import OptionParser
            self.oOpt=OptionParser(sUsage,version="%prog "+sVer)
            self.iMode=0
        self.__initDftArg__()
    def __initDftArg__(self):
        """
        ### parameter
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            # ----- end:
            # +++++ beg:initialize
            iRet=0
            # ----- end:initialize
            # +++++ beg:finalize
            if self.iMode==1:
                self.oOpt.add_argument("-v", type=int, nargs='?' , 
                    default=0,
                    dest="iVerbose", 
                    )
            else:
                self.oOpt.add_option("-v", action="store_true", dest="verbose", default=True)
            # ----- end:finalize
            return iRet
        except:
            traceback.print_exc()
            return -1
    def addOpt(self,sVar,sDft='',sHlp='',sMeta='',sVerbose=None,
                funcVld=None,**kwargs):
        """add option to command line parser

        variable name define command line argument and type,
        applying rules establish standard and reduce effort.
        information made available to implement GUI later.

        For example sVar sCfgFN specify string argument --cfgFN,
        a file name browser shall be used as GUI control.

        ### parameter
            sVar    ... option name
                first letter define type
                    i   ... int
                    r   ... float
                    s   ... string
                following letters build argument
                suffix define GUI control
                    FN  ... file name browser
                    DN  ... directory name browser
                    Chc ... choice
                    Int ... integer
                    Num ... numerical
            sDft    ... default value
            sHlp    ... help string
            sMeta   ... variable example
            sVerbose .. verbose label
                if set, shown automatically if verbose activated
            funcVld ... validation function (callable)
            kwargs  ... keyword arguments passed to funcVld
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sArg=''.join(['--',sVar[1].lower(),sVar[2:]])
            cType=sVar[0]
            # ----- end:initialize
            # +++++ beg:store argument
            if self.iVerbose>0:
                print('addOpt')
                sVrbLv='   '
            if cType in self.MAP_TYPE:
                pass
            else:
                cType='s'       # known type changed to string
            self.lArg.append(sVar)
            self.dArg[sVar]={
                'sArg':sArg,
                'cType':cType,
                'sDft':sDft,
                'sHlp':sHlp,
                'sVar':sVar,
                'sVerbose':sVerbose,
                'funcVld':funcVld,
                'kwargs':kwargs,
                }
            if self.iVerbose>0:
                print(sVrbLv,'sVar',sVar)
                print(sVrbLv,'d:%r'%(self.dArg[sVar]))
            # ----- end:store argument
            # +++++ beg:add to command line parser
            if self.iMode==1:
                self.oOpt.add_argument(
                        sArg,
                        type=self.MAP_TYPE[cType],
                        default=sDft,
                        help=sHlp,metavar=sMeta,
                        dest=sVar
                        )
                return 1
            else:
                self.oOpt.add_option('',sArg,
                        dest=sVar,
                        default=sDft,
                        help=sHlp,
                        metavar=sMeta,
                        )
                return 1
            # ----- end:add to command line parser
            return 0
        except:
            traceback.print_exc()
            return -1
    def prcParse(self,args):
        """
        ### parameter
            args    ... tuple of command line arguments
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            # ----- end:
            # +++++ beg:initialize
            iRet=0
            # ----- end:initialize
            # +++++ beg:parse
            if self.iMode==1:
                self.oOptRes=self.oOpt.parse_args(args)
                if self.oOptRes.iVerbose<=0:
                    self.iVerbose=0
                else:
                    self.iVerbose=self.oOptRes.iVerbose
            else:
                (self.oOptRes,args)=self.oOpt.parse_args(args=args)
                if self.oOptRes.verbose==False:
                    self.iVerbose=0
            # ----- end:parse
            # +++++ beg:verbose
            if self.iVerbose>0:
                print('self.oOptRes:%r'%(self.oOptRes))
                #print('  len:%d'%(len(self.oOptRes)))
            # ----- end:verbose
            # +++++ beg:finalize
            for sVar in self.lArg:
                try:
                    dDef=self.dArg[sVar]
                    oVal=getattr(self.oOptRes,sVar,dDef.get('sDft',''))
                    if oVal=='None':
                        oVal=None
                    # +++++ beg:validate
                    funcVld=dDef.get('funcVld',None)
                    if funcVld is not None:
                        kwargs=dDef.get('kwargs',{})
                        oVal=funcVld(oVal,**kwargs)
                    # ----- end:validate
                    setattr(self,sVar,oVal)
                    if self.iVerbose>0:
                        sVerbose=dDef.get('sVerbose',None)
                        if  sVerbose is not None:
                            try:
                                cType=dDef.get('cType','s')
                                if cType in ['s']:
                                    print('%20s:%r'%(sVerbose,oVal))
                                elif cType in ['i']:
                                    print('%20s:%d'%(sVerbose,oVal))
                                elif cType in ['r']:
                                    print('%20s:%f'%(sVerbose,oVal))
                                else:
                                    print('%20s:%r'%(sVerbose,oVal))
                            except:
                                traceback.print_exc()
                    iRet+=1
                except:
                    traceback.print_exc()
            # ----- end:finalize
            return iRet
        except:
            traceback.print_exc()
            return -1
    def __initPan__(self,oPar,wNb):
        """
        ### parameter
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            # ----- end:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmArg::__initPan__'
            oPar.logDbg('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            # ----- end:initialize
            # +++++ beg:
            lWid=[]
            iBrowser=0
            for sArg in self.lArg:
                if sArg.endswith('DN'):
                    iBrowser=1
                elif sArg.endswith('FN'):
                    iBrowser=1
            if iBrowser>0:
                iCol=3
            else:
                iCol=2
            for sArg in self.lArg:
                dDef=self.dArg[sArg]
                sArg=dDef['sArg'][2:]
                sVar=dDef['sVar'][1:]
                lWid.append(['lbl','lbl'+sVar,sArg])
                if sArg.endswith('Chc'):
                    sVal=getattr(self,dDef['sVar'])
                    sHlp=dDef['sHlp']
                    iPos=sHlp.find(';')
                    if iPos>0:
                        lChc=sHlp[iPos+1:].split('|')
                    else:
                        lChc=['???','---']
                    tVal=[sVal,lChc]
                    lWid.append(['chc','chc'+sVar[:-3],tVal])
                    lWid.append([None])
                elif sArg.endswith('Int'):
                    sVal=getattr(self,dDef['sVar'])
                    sHlp=dDef['sHlp']
                    iPos=sHlp.find(';')
                    if iPos>0:
                        lLmt=[s.strip() for s in sHlp[iPos+1:].split('<=')]
                    else:
                        lLmt=['0','x','100']
                    tVal=[sVal,lLmt]
                    lWid.append(['spn','spn'+sVar,tVal])
                    lWid.append([None])
                elif sArg.endswith('Num'):
                    lWid.append([None])
                else:
                    lWid.append(['txt','txt'+sVar,getattr(self,dDef['sVar'])])
                    if iBrowser>0:
                        if sArg.endswith('DN'):
                            lWid.append(['cbd','cb'+sVar,'txt'+sVar])
                        elif sArg.endswith('FN'):
                            sLnk=dDef['sVar'][:-2]+'DN'
                            if sLnk in self.dArg:
                                # +++++ beg:linked variable found
                                sWidLnk='txt'+sLnk[1:]
                                lWid.append(['cbf','cb'+sVar,'txt'+sVar,sWidLnk])
                                # ----- end:linked variable found
                            else:
                                lWid.append(['cbf','cb'+sVar,'txt'+sVar])
                        else:
                            lWid.append([None])
            from lindworm.ldmWidPanel import ldmWidPanelFlxGrd
            wPan=ldmWidPanelFlxGrd(sLogger='ldmArg',iLv=0,
                        parent=oPar,
                        iCol=iCol,lCol=[1],
                        lWid=lWid)
            oPar.pnCLI=wPan
            #oWid=ldmWidSizeRspWid(wNb, -1,mgr=self.mgrMain)
            wNb.AddPage(wPan.GetWid(),'CLI',True)
            # ----- end:
            # +++++ beg:finalize
            oPar.logDbg('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            oPar.logTB()
            return -1
    def GetVerbose(self,iVerbose=-1):
        """get verbose level 
        ### parameter
        ### return
            True    ... object verbose >= iVerbose
            False   ... object verbose < iVerbose
            number  ... object verbose if iVerbose < 0
        """
        if iVerbose>=0:
            if self.iVerbose>=iVerbose:
                return True
            else:
                return False
        else:
            return self.iVerbose
    def GetVerboseDbg(self,iVerbose=-1):
        """get verbose level in debugging mode
        ### parameter
        ### return
            True    ... object verbose >= iVerbose
            False   ... object verbose < iVerbose
                    ... log level not in debugging
            number  ... object verbose if iVerbose < 0
        """
        if self.iIsDbg>0:
            return self.GetVerbose(iVerbose=iVerbose)
        else:
            return False



#----------------------------------------------------------------------------
# Name:         lndStorage.py
# Purpose:      storage base class
#
# Author:       Walter Obweger
#
# Created:      20200104
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import os
import logging
import traceback
import json

from lindworm.logUtil import logGet
from lindworm.logUtil import ldmUtilLog
import lindworm.ldmOS as ldmOS

class ldmStorage(ldmUtilLog):
    def __init__(self,iModeRev=0,sLogger='',iLv=1,iVerbose=0):
        """constructor
        ### parameter
            iModeRev    ... reverse processing
                    1   ... FILO, first in last out
                    0   ... FIFO, first in first out
            sLogger     ... log origin
            iLv         ... log level short
            iVerbose    ... higher values add more logs
        """
        ldmUtilLog.__init__(self,sLogger=sLogger,iLv=iLv,
                            iVerbose=iVerbose)
        #self.iVerbose=iVerbose
        #self.oLog=logGet(sLogger,iLevel)
        self.__initCfg__()
        self.__initDef__()
        self.__initDat__()
        self.__initPrc__()
        self.clrEnd()
        self.dCfgDft['iModeRev']=iModeRev
    def __initCfg__(self):
        """initialize configuration properties
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::__initCfg__'
            self.oLog.debug('beg:%s'%(sOrg))
            # ----- end:initialize
            # +++++ beg:initialize configuration attributes
            self.sCfgFN=''
            self.dCfg=None
            self.dCfgDft={
                'lAtr':['dDat']
                }
            # ----- end:initialize configuration attributes
            self.oLog.debug('end:%s'%(sOrg))
            return 1
        except:
            self.logTB()
            return -1
    def __initDef__(self):
        """initialize definition properties
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::__initDef__'
            self.oLog.debug('beg:%s'%(sOrg))
            # ----- end:initialize
            # +++++ beg:initialize data
            self.lDef=[]        # list definition been processed
            self.dArg={}        # arguments data
            self.dRef={}        # reference data
            # ----- end:initialize data
            self.oLog.debug('end:%s'%(sOrg))
            return 1
        except:
            self.logTB()
            return -1
    def __initDat__(self):
        """initialize data properties
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::__initDat__'
            self.oLog.debug('beg:%s'%(sOrg))
            # ----- end:initialize
            # +++++ beg:initialize data
            self.dDat={}        # dictionary processed files, performance
            # ----- end:initialize data
            self.oLog.debug('end:%s'%(sOrg))
            return 1
        except:
            self.logTB()
            return -1
    def __initPrc__(self):
        """initialize processing properties
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::__initPrc__'
            self.oLog.debug('beg:%s'%(sOrg))
            # ----- end:initialize
            # +++++ beg:initialize data
            self.iAct=-1                    # dictionary processed files, performance
            # ----- end:initialize data
            self.oLog.debug('end:%s'%(sOrg))
            return 1
        except:
            self.logTB()
            return -1
    def clrAll(self):
        """clear data properties
        """
        self.clrDef()
        self.clrDat()
        self.clrEnd()
        self.__initPrc__()
    def clrDef(self):
        """clear stored file names to process
        """
        iRet=len(self.lDef)
        self.lDef=[]
        return iRet
    def clrDat(self):
        """clear data properties
        """
        self.__initDat__()
    def clrEnd(self):
        """clear stored file names to process
        """
        self.sDefEnd=None
        self.oArgEnd=None
        self.oDatEnd=None
        self.oRefEnd=None
        return 1
    def getLstDef(self):
        """get definition list
        ### return
            []  ... list of definition stored
        """
        try:
            return self.lDef
        except:
            self.logTB()
            return []
    def getDef(self):
        """get current file name
        ### return
            iRet    ... return code
                >0  ... okay processing done
                =0  ... okay nop
                <0  ... error
            sFN     ... file name : str
        """
        try:
            if len(self.lDef)>0:
                if self.getCfg('iModeRev',sType='int',oDft=0)==1:
                    sDef=self.lDef[-1]
                else:
                    sDef=self.lDef[0]
                return 1,sDef
            return 0,None
        except:
            self.logTB()
            return -1,None
    def getDat(self,sDef=None):
        """current data object
        ### parameter
            sDef    ... definition to search, None = get current reference
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            if sDef is None:
                iRet,sDef=self.getDef()
            else:
                iRet=1
            if iRet>0:
                if sDef in self.dDat:
                    oDat=self.dDat[sDef]
                    return 1,oDat
            return 0,None
        except:
            self.logTB()
            return -1,None
    def getRef(self,sDef=None):
        """get current reference object
        ### parameter
            sDef    ... definition to search, None = get current reference
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            if sDef is None:
                iRet,sDef=self.getDef()
            else:
                iRet=1
            if iRet>0:
                if sDef in self.dDat:
                    oRef=self.dRef[sDef]
                    return 1,oRef
            return 0,None
        except:
            self.logTB()
            return -1,None
    def prcIsDone(self,sDef):
        """check definition has been queues/processed already  
        ### parameter
            sDef ... definition to check
        ### return
            >0  ... okay marked for processing done
            =0  ... okay unknown file name
            <0  ... error
        """
        try:
            if sDef in self.dDat:
                iRet=1
            else:
                iRet=0
            self.oLog.debug('%s sDef:%s iRet:%d'%('ldmStorage::prcIsDone',
                            sDef,iRet))
            return iRet
        except:
            self.logTB()
            return -1
    def bldDat(self,sDef,oRef=None,**kwargs):
        """check definition has been queues/processed already  
        ### parameter
            sDef    ... definition to check
            oRef    ... reference object
            kwargs  ... flexible keyword argument
        ### return
            oDat    ... okay marked for processing done
        """
        try:
            sOrg='ldmStorage::bldDat'
            if self.iVerbose>5:
                self.oLog.debug('%s sDef:%s oRef:%r kwargs:%r'%(sOrg,sDef,oRef,kwargs))
            elif self.iVerbose>0:
                self.oLog.debug('%s sDef:%s'%(sOrg,oRef))
            return {}
        except:
            self.logTB()
            return None
    def prcBeg(self,sDef,oRef=None,**kwargs):
        """processing begin, add file name and objects to stack
        ### parameter
            sDef    ... definition name : str
            oDat    ... data object
            oRef    ... reference object
            kwargs  ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            sOrg='ldmStorage::prcBeg'
            # ----- end:initialize
            # +++++ beg:add file name and auxillary objects to storage
            self.oLog.debug('beg:%s sDef:%s'%(sOrg,sDef))
            if self.prcIsDone(sDef)==0:
                self.lDef.append(sDef)
                self.dArg[sDef]=kwargs
                self.dRef[sDef]=oRef
                # +++++ beg:build data object
                oDat=self.bldDat(sDef,oRef=oRef,**kwargs)
                self.dDat[sDef]=oDat
                # ----- end:build data object
                iRet=1
            self.oLog.debug('end:%s iRet:%d'%(sOrg,iRet))
            # ----- end:add file name and auxillary objects to storage
            return iRet
        except:
            self.logTB()
            return -1
    def prcExc(self,**kwargs):
        """processing end
        ### parameter
            kwargs  ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iRet=0
            # ----- end:initialize
            # +++++ beg:
            if len(self.lDef)<1:
                self.oLog.debug('skp:%s iRet:%d empty stack'%('ldmStorage::prcExc',iRet))
                return iRet
            iR,sDef=self.getDef()
            self.oLog.debug('beg:%s sDef:%s'%('ldmStorage::prcExc',sDef))
            if self.iVerbose>0:
                self.oLog.debug('    kwargs:%r'%(kwargs))
            self.oLog.debug('end:%s iRet:%d'%('ldmStorage::prcExc',iRet))
            # ----- end:
            return iRet
        except:
            self.logTB()
            return -1
    def prcEnd(self,**kwargs):
        """processing end
        ### parameter
            **kwargs ... flexible keyword argument
        ### return
            iRet    ... return code
                >0  ... okay processing done
                =0  ... okay nop
                <0  ... error
            sFN     ... filename
            oDat    ... data object
            oRef    ... reference object
        """
        try:
            # +++++ beg:initialize
            iRet=0
            self.clrEnd()
            # ----- end:initialize
            # +++++ beg:
            iCntDef=len(self.lDef)
            if iCntDef>0:
                if self.getCfg('iModeRev',sType='int',oDft=0)==1:
                    self.sDefEnd=self.lDef.pop()
                else:
                    self.sDefEnd=self.lDef[0]
                    del self.lDef[0]
                self.oArgEnd=self.dArg.get(self.sDefEnd,None)
                self.oDatEnd=self.dDat.get(self.sDefEnd,None)
                self.oRefEnd=self.dRef.get(self.sDefEnd,None)
                iRet=1
            # ----- end:
            return iRet
        except:
            self.logTB()
            return -1
    def saveJson(self,oDat,sFN,sDN=None,sSfx=None):
        """write object attribute data to json file
        ### parameter
            sFN     ... file name
            sDN     ... directory
            sSfx    ... suffix
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::saveJson'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            iRet=0
            # ----- end:initialize
            # +++++ beg:build file name
            iR,sFullFN=ldmOS.bldFN(sFN,sDN=sDN,
                                sExt='json',sSfx=sSfx,
                                oLog=self.oLog,iVerbose=1)
            self.oLog.debug('   :iR:%d sFullFN:%r'%(iR,sFullFN))
            if iR<=0:
                return 0
            iRet=1
            # ----- end:build file name
            # +++++ beg:save json
            with open(sFullFN,'w') as oFile:
                sDat=json.dumps(oDat)
                oFile.write(sDat)
            # ----- end:save json
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def loadJson(self,dDat,sK,sFN,sDN=None,iUpdate=0):
        """load directory list to json
        ### parameter
            dDat    ... dictionary
            sK      ... key to place loaded data
            sFN     ... file name
            sDN     ... optional directory
            iUpdate ... update dDat[sK]
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:build input file name
            sOrg='ldmStorage::loadJson'
            self.logDbg('beg:%s sK:%r sFN:%r sDN:%r iUpdate:%d',
                        sOrg,sK,sFN,sDN,iUpdate)
            iRet=0
            oVal=None
            if sDN is not None:
                sTmpFN=os.path.join(sDN,sFN)
            else:
                sTmpFN=sFN
            # ----- end:build input file name
            # +++++ beg:load info to json
            with open(sTmpFN,'r') as oFile:
                sDat=oFile.read()
                oVal=json.loads(sDat)
                if iUpdate>0:
                    if sK in dDat:
                        dDat[sK].update(oVal)
                    else:
                        dDat[sK]=oVal
                else:
                    dDat[sK]=oVal
            iRet=1
            # ----- end:load info to json
            self.logDbg('end:%s iRet:%d',sOrg,iRet)
            return iRet
        except:
            self.logTB()
            return -1
    def saveAtr(self,sFN,sDN=None,lAtr=None):
        """save object attribute data to json file
        ### parameter
            sFN     ... file name
            sDN     ... directory
            lAtr    ... attribute list
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::saveAtr'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            iRet=0
            # ----- end:initialize
            # +++++ beg:finalize
            if lAtr is None:
                iR,lAtr=self.getCfg('lAtr')
                if iR<=0:
                    self.oLog.error('   :getCfg problem iR:%d'%(iR))
                    lAtr=['dDat']
            for sAtr in lAtr:
                if sAtr in self.__dict__:
                    oAtr=self.__dict__[sAtr]
                    iR=self.saveJson(oAtr,sFN,sDN,sSfx=sAtr)
                    if iR>0:
                        iRet+=1
            # ----- end:read cfg file
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def saveDat(self,sFN,sDN=None,lKey=None,sAtr='dDat'):
        """save object data to json file
        ### parameter
            sFN     ... file name
            sDN     ... directory
            lKey    ... keys
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::saveDat'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            iRet=0
            # ----- end:initialize
            # +++++ beg:finalize
            if self.iVerbose>0:
                lAtr=list(self.__dict__.keys())
                self.oLog.debug('    sAtr:%s lAtr:%r lKey:%r'%(sAtr,lAtr,lKey))
            
            if sAtr in self.__dict__:
                if self.iVerbose>0:
                    self.oLog.debug('    sAtr:%s found lKey:%r'%(sAtr,lKey))
                oAtr=self.__dict__[sAtr]
                if lKey is not None:
                    iLenKey=len(lKey)
                    for sK in lKey:
                        oVal=oAtr[sK]
                        if iLenKey>1:
                            sSfx='.'.join([sAtr,sK])
                        else:
                            sSfx=None
                        iR=self.saveJson(oVal,sFN,sDN,sSfx=sSfx)
                        if iR>0:
                            iRet+=1
                else:
                    iR=self.saveJson(oAtr,sFN,sDN,sSfx=None)
                    if iR>0:
                        iRet+=1
            # ----- end:read cfg file
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def loadDat(self,sFN,sDN=None,sKey=None,sAtr='dDat'):
        """load object data to json file
        ### parameter
            sFN     ... file name
            sDN     ... directory
            lKey    ... keys
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::loadDat'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            iRet=0
            # ----- end:initialize
            # +++++ beg:finalize
            if self.iVerbose>0:
                lAtr=list(self.__dict__.keys())
                self.oLog.debug('    sAtr:%s lAtr:%r sKey:%r'%(sAtr,lAtr,sKey))
            
            if sAtr in self.__dict__:
                if self.iVerbose>0:
                    self.oLog.debug('    sAtr:%s found sKey:%r'%(sAtr,sKey))
                oAtr=self.__dict__[sAtr]
                if sKey is not None:
                    iR=self.loadJson(oAtr,sKey,sFN,sDN)
                    if iR>0:
                        iRet+=1
                else:
                    iR=self.loadJson(self.__dict__,sAtr,sFN,sDN,sSfx=None)
                    if iR>0:
                        iRet+=1
            else:
                iR=self.loadJson(self.__dict__,sAtr,sFN,sDN,sSfx=None)
                if iR>0:
                    iRet+=1
                        
            # ----- end:read cfg file
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def getCfg(self,sKey,sType=None,oDft=None):
        """get configuration value
        ### parameter
            sKey    ... configuration key
            sType   ... type enforcement
            oDft    ... default value
        ### return
            return code
                >0  ... okay processing done
                =0  ... okay nop
                <0  ... error
            value
        """
        try:
            # +++++ beg:initialize
            iRet=0
            oVal=oDft
            # ----- end:initialize
            # +++++ beg:find value
            if self.dCfg is not None:
                if sKey in self.dCfg:
                    iRet=2
                    oVal=self.dCfg[sKey]
            # ----- end:find value
            # +++++ beg:find default value
            if iRet==0:
                if self.dCfgDft is not None:
                    if sKey in self.dCfgDft:
                        iRet=1
                        oVal=self.dCfgDft[sKey]
            # ----- end:find default value
            # +++++ beg:type enforcement
            if iRet>0:
                if sType=='int':
                    try:
                        iVal=int(oVal)
                        return iRet,iVal
                    except:
                        iVal=oDft
                        return iRet,iVal
            # ----- end:type enforcement
            return iRet,oVal
        except:
            self.logTB()
            return -1,None
    def loadCfg(self,sCfgFN='cfg.json'):
        """read configuration file, given in json format
        ### parameter
            sCfgDN  ... configuration file name
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmStorage::loadCfg'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            iRet=0
            # ----- end:initialize
            # +++++ beg:read cfg file
            self.sCfgFN=sCfgFN
            self.dCfg=None
            if os.path.exists(self.sCfgFN):
                with open(self.sCfgFN,'r') as oFile:
                    self.dCfg=json.loads(oFile.read())
                iRet=1
            # ----- end:read cfg file
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def __tpl(self):
        """
        ### parameter
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            # ----- end:
            # +++++ beg:initialize
            iRet=0
            sOrg='::'
            self.oLog.debug('beg:%s iVerbose:%d'%(sOrg,self.iVerbose))
            # ----- end:initialize
            # +++++ beg:finalize
            self.oLog.debug('end:%s  iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1



#----------------------------------------------------------------------------
# Name:         lndMD.py
# Purpose:      MD class
#
# Author:       Walter Obweger
#
# Created:      20200108
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import logging
import traceback
import os.path

from lindworm.ldmStorageLine import ldmStorageLine
from lindworm.ldmStr import ldmStr

class ldmMD(ldmStorageLine):
    SEP_HDR='.'
    def __init__(self,sFN,sLogger='',iLv=1,
                 iSkipDN=-1,iVerbose=0):
        """constructor
        ### parameter
            iModeRev    ... reverse processing
                    1   ... FILO, first in last out
                    0   ... FIFO, first in first out
            sLogger     ... log origin
            iLv         ... logging level
            iVerbose    ... higher values add more logs
        """
        ldmStorageLine.__init__(self,
                    iModeRev=0,
                    sLogger=sLogger,
                    iLv=iLv,
                    iVerbose=iVerbose)
        self.sFullFN=sFN
        self.oStr=ldmStr()
        sTmpDN,sTmpFN=os.path.split(sFN)
        self.sDN=sTmpDN
        self.sFN=sTmpFN
        self.iSkipDN=iSkipDN
    def __initDat__(self):
        """initialize data properties
        """
        try:
            sOrg='ldmStorage::__initDat__'
            self.oLog.debug('beg:%s'%(sOrg))
            # +++++ beg:initialize parent
            ldmStorageLine.__initDat__(self)
            # ----- end:initialize parent
            # +++++ beg:initialize data
            self.lHdr=[]        # list header per level during processing
            self.dHdr={}        # dictionary header 
            self.lHdrNavLv=[]   # list header navigation and level
            # ----- end:initialize data
            self.oLog.debug('end:%s'%(sOrg))
            return 1
        except:
            self.logTB()
            return -1
    #def clr(self):
    #    """clear data properties
    #    """
    #    self.__initDat__()
    def getHdrNavLv(self,iOfs=0):
        """get header level, maximal 6 allowed
        ### parameter
            iOfs    ... offset to header
        ### return
            sNav    ... navigation data
            sHdr    ... header
            iLv     ... level, 1 to 6, -1= out of range
        """
        try:
            if len(self.lHdrNavLv)>iOfs:
                return self.lHdrNavLv[iOfs]
            return '','',-1
        except:
            self.logTB()
            return '','',-1
    def getHdrLv(self,sLine):
        """get header level, maximal 6 allowed
        ### return
            >0  ... okay header found
            =0  ... okay nop
            <0  ... error
        """
        try:
            sTmp=sLine.strip()
            iLen=min(6,len(sTmp))
            for iOfs in range(iLen):
                if sTmp[iOfs]!='#':
                    return iOfs
            return 0
        except:
            self.logTB()
            return -1
    def bldDat(self,sDef,oRef=None,**kwargs):
        """check definition has been queues/processed already  
        ### parameter
            sDef    ... definition to check
            oRef    ... reference object
            kwargs  ... flexible keyword argument
        ### return
            oDat    ... okay marked for processing done
        """
        try:
            sOrg='ldmMD::bldDat'
            if self.iVerbose>0:
                self.oLog.debug('beg:%s sDef:%s oRef:%r kwargs:%r'%(sOrg,sDef,oRef,kwargs))
            oDat=kwargs.get('oDat',self)
            if self.iVerbose>0:
                self.oLog.debug('end:%s sDef:%s oDat:%r'%(sOrg,sDef,oDat))
            return oDat
        except:
            self.logTB()
            return None
    def handleLink(self,sLink,oGtrMD=None):
        """handle link
        ### parameter
            sLink   ... link data
            oGtrMD  ... global storage
        ### return
            ''      ... changed string
            None    ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmMD::handleLink'
            iVerboseLmt=5
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('beg:%s sLink:%s'%(sOrg,sLink))
            iRet=0
            sNav=None
            oMD=None
            # ----- end:initialize
            # +++++ beg:
            sTmpFN=sLink.lower()
            if sTmpFN.endswith('.md'):
                sFullFN=os.path.join(self.sDN,sLink)
                if os.path.exists(sFullFN)==False:
                    self.oLog.warning("    handleLink return -5 sFullFN:%s file does not exist, skip"%(sFullFN))
                    iRet=-5
                else:
                    sAbsFullFN=os.path.abspath(sFullFN)
                    oMD=ldmMD(sAbsFullFN,
                            sLogger='oMD',
                            iSkipDN=self.iSkipDN,
                            iVerbose=self.iVerbose)
                    if oGtrMD is not None:
                        iRetDone=oGtrMD.prcIsDone(sAbsFullFN)
                        if iRetDone==0:
                            oGtrMD.prcBeg(sAbsFullFN,oDat=oMD,oRef=None)
                        elif iRetDone>0:
                            iRetMD,oMD=oGtrMD.getDat(sAbsFullFN)
                        else:
                            oMD=None
                    else:
                        iRetDone=self.prcIsDone(sAbsFullFN)
                        if iRetDone>0:
                            iRetMD,oMD=self.getDat(sAbsFullFN)
                    if iRetDone==0:
                        iRet=self.prcBeg(sAbsFullFN,oDat=oMD,oRef=None)
                        if iRet>0:
                            # +++++ beg:process sub file
                            oMD.prcBeg(sAbsFullFN,oDat=oMD,oRef=None)
                            oMD.prcExc(oGtrMD=oGtrMD)
                            oMD.prcEnd()
                            # ----- end:process sub file
                    if iRetDone>=0:
                        # +++++ beg:process sub file
                        sNav,sHdr,iLv=oMD.getHdrNavLv()
                        if self.iVerbose>iVerboseLmt:
                            self.oLog.debug('    %s sNav:%r sHdr:%r iLv:%d'%(sOrg,sNav,sHdr,iLv))
                        # ----- end:process sub file
            # ----- end:
            # +++++ beg:finalize
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('end:%s iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return sNav
        except:
            self.logTB()
            return None
    def getHdrPrefix(self):
        """get header prefix, reflect origin
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:find file extension
            iPosExt=self.sFN.rfind('.')
            if iPosExt<0:
                iPosExt=len(self.sFN)
                sFN=self.sFN[:iPosExt]
            else:
                sFN=self.sFN
            # ----- end:find file extension
            # +++++ beg:
            sPart1=self.sDN+self.SEP_HDR+sFN
            sPart1=self.oStr.replaceIgnore(sPart1[self.iSkipDN:])
            # ----- end:
            return sPart1
        except:
            self.logTB()
            return ''
    def getHdrIdf(self,sHdr):
        """get header path as string separated by SEP_HDR
        ### parameter
            sHdr    ... header
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iVerboseLmt=10
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('beg: sHdr:%s'%(sHdr))
            iRet=0
            # ----- end:initialize
            # +++++ beg:
            if sHdr is None:
                sHdr=self.SEP_HDR.join(self.lHdr)
            #else:
            #    sHdr=sHdr.strip()
            sHdr=self.oStr.replaceIgnore(sHdr)
            # ----- end:
            # +++++ beg:finalize
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('end: sHdr:%s'%(sHdr))
            # ----- end:finalize
            return sHdr
        except:
            self.logTB()
            return ''
    def setHdr(self,iHdr,sHdr):
        """store heading at proper header location
        ### parameter
            iHdr    ... header level
            sHdr    ... header
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iVerboseLmt=10
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('beg: iHdr:%d sHdr:%s'%(iHdr,sHdr))
            iRet=0
            # ----- end:initialize
            # +++++ beg:update header stack
            iLenHdr=len(self.lHdr)
            if iLenHdr>=iHdr:
                self.lHdr=self.lHdr[:iHdr]
            else:
                for iOfs in range(iLenHdr,iHdr):
                    self.lHdr.append('')
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('     iHdr:%d %d lHdr:%r'%(iHdr,
                                iLenHdr,self.lHdr))
            self.lHdr[iHdr-1]=sHdr
            iRet=iHdr
            # ----- end:update header stack
            # +++++ beg:finalize
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('end: lHdr:%r'%(self.lHdr))
            # ----- end:finalize
            return iRet
        except:
            self.logTB()
            return -1
    def prcLine(self,iLine,sLine,**kwargs):
        """add line to storage
        ### parameter
            iLine   ... line number : int
            sLine   ... line : str
            oGtrMD   ... gather object
            **kwargs ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmMD::prcLine'
            iVerboseLmt=0
            iRet=0
            iPosBeg=0
            iIsHdr=0
            iIsLink=0
            iIsImage=0
            iPosLblBeg=-1
            iPosLblEnd=-1
            iPosDatBeg=-1
            iPosDatEnd=-1
            iLen=len(sLine)
            iPosLF=sLine.rfind('\n')
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('beg:%s iLine:%6d sLine:%s'%(sOrg,iLine,sLine[:iPosLF]))
            oGtrMD=kwargs.get('oGtrMD',None)
            # ----- end:initialize
            # +++++ beg:check header
            iHdr=self.getHdrLv(sLine)
            if iHdr>0:
                # +++++ beg:
                iPosLblBeg=iHdr+1
                iPosLblBeg=self.oStr.findWhiteSpace(sLine,iLen,iPosLblBeg,iMode=0)
                iPosLblEnd=self.oStr.findWhiteSpace(sLine,iLen,iPosLblBeg,iMode=1)
                if self.iVerbose>iVerboseLmt:
                    self.oLog.debug('    iHdr:%d iPosLblBeg:%d iPosLblEnd:%d'%(iHdr,iPosLblBeg,iPosLblEnd))
                iPosDatBeg,iPosDatEnd=self.oStr.findBegEnd(sLine,iLen,
                                                    iPosBeg,'{','}')
                if self.iVerbose>iVerboseLmt:
                    self.oLog.debug('    iHdr:%d iPosDatBeg:%d iPosDatEnd:%d'%(iHdr,iPosDatBeg,iPosDatEnd))
                if iPosDatBeg<0:
                    iPosLblEnd=sLine.rfind('\r')
                    if iPosLblEnd<0:
                        iPosLblEnd=iPosLF
                if iPosLblBeg>0:
                    #if iPosLblEnd<0:
                    #    iPosLblEnd=sLine.rfind('\r')
                    #    if iPosLblEnd<0:
                    #        iPosLblEnd=iPosLF
                    if iPosLblEnd>0:
                        sHdr=sLine[iPosLblBeg:iPosLblEnd]
                self.setHdr(iHdr,sHdr)
                # ----- end:
                # +++++ beg:
                #iPosDatBeg=iHdr+1
                if self.iVerbose>iVerboseLmt:
                    self.oLog.debug('    iHdr:%d sHdr:%s'%(iHdr,sHdr))
                if iPosDatBeg>0:
                    iPosLblEnd=iPosDatBeg-1
                    if iPosDatEnd>0:
                        iPosDatBeg=iPosDatBeg+1
                        sTmpDat=sLine[iPosDatBeg:iPosDatEnd]
                        sHdrNav='{#%s}'%(sTmpDat)
                        self.lHdrNavLv.append((sHdrNav[1:-1],sHdr,iHdr))
                    else:
                        self.lHdrNavLv.append((sHdr,sHdr,iHdr))
                else:
                    iPosLblEnd=sLine.rfind('\r')
                    if iPosLblEnd<0:
                        iPosLblEnd=iPosLF
                    sTmpDat=None
                    #sHdr=sLine[iPosDatBeg:iPosLblEnd]
                    sHdrNav='{#%s%s%s}'%(self.getHdrPrefix(),self.SEP_HDR,self.getHdrIdf(sTmpDat))
                    sLine=sLine[:iPosLblEnd]+sHdrNav+sLine[iPosLF:]
                    self.lHdrNavLv.append((sHdrNav[1:-1],sHdr,iHdr))
                iPosBeg=-1
                iIsHdr=1
                # ----- end:
            # ----- end:check header
            # +++++ beg:check link
            while iPosBeg>=0:
                # +++++ beg:clear loop variables
                iIsLink=0
                iIsImage=0
                iPosLblBeg=-1
                iPosLblEnd=-1
                iPosDatBeg=-1
                iPosDatEnd=-1
                # ----- end:clear loop variables
                # +++++ beg:find link start
                iPosLblBeg,iPosLblEnd=self.oStr.findBegEnd(sLine,iLen,
                                                    iPosBeg,'[',']')
                #iPosLblBeg=sLine.find('[',iPosBeg)
                if self.iVerbose>iVerboseLmt:
                    self.oLog.debug('  iPosBeg:%d iPosLblBeg:%d iPosLblEnd:%d '%(iPosBeg,iPosLblBeg,iPosLblEnd))
                if iPosLblBeg>0:
                    if sLine[iPosLblBeg-1]=='!':
                        iIsImage=1
                    else:
                        iIsImage=0
                    if self.iVerbose>iVerboseLmt:
                        self.oLog.debug('  lnk beg found iPosLblBeg:%d iPosLblEnd:%d '
                                        'iIsImage:%d'%(iPosLblBeg,iPosLblEnd,iIsImage))
                # ----- end:find link start
                if iPosLblBeg>=0:
                    # +++++ beg:find link parts
                    #iPosLblEnd=sLine.find(']',iPosLblBeg)
                    if iPosLblEnd>0:
                        iPosDatBeg,iPosDatEnd=self.oStr.findBegEnd(sLine,iLen,
                                                    iPosLblEnd,'(',')')
                        if self.iVerbose>iVerboseLmt:
                            self.oLog.debug('  lnk beg found iPosDatBeg:%d iPosDatEnd:%d '
                                            'iIsImage:%d'%(iPosDatBeg,iPosDatEnd,iIsImage))
                        #iPosDatBeg=sLine.find('(',iPosLblEnd)
                        if iPosDatBeg>0:
                            #iPosDatEnd=sLine.find(')',iPosDatBeg)
                            if iPosDatEnd>0:
                                # found
                                iIsLink=1
                            else:
                                iPosBeg=-1
                        else:
                            if iLen>(iPosLblEnd+2):
                                if sLine[iPosLblEnd+1]==':':
                                    # +++++ beg:find second link part
                                    iPosBeg=-1
                                    iPosDatBeg=self.oStr.findWhiteSpace(sLine,iLen,
                                                        iPosLblEnd+2,
                                                        iMode=0)
                                    if iPosDatBeg>0:
                                        iPosDatEnd=self.oStr.findWhiteSpace(sLine,iLen,
                                                        iPosDatBeg,
                                                        iMode=1)
                                        if iPosDatEnd>0:
                                            # found
                                            iIsLink=1
                                        else:
                                            iPosBeg=-1
                                    else:
                                        iPosBeg=-1
                                    # ----- end:find second link part
                                else:
                                    iPosBeg=-1
                            else:
                                iPosBeg=-1
                    else:
                        iPosBeg=-1
                    # ----- end:find link parts
                else:
                    iPosBeg=-1
                # ----- end:check link
                # +++++ beg:process link
                if self.iVerbose>20:
                    self.oLog.debug('  iPosBeg:%d iIsLink:%d iIsImage:%d'
                                    ''%(iPosBeg,iIsLink,iIsImage))
                if iIsLink>0:
                    # +++++ beg:link info
                    if self.iVerbose>20:
                        self.oLog.debug('  iPosBeg:%d iPosLblBeg:%d iPosLblEnd:%d '
                                    'iPosDatBeg:%d iPosDatEnd:%d'%(iPosBeg,iPosLblBeg,iPosLblEnd,
                                    iPosDatBeg,iPosDatEnd))
                    iPosLblBeg=iPosLblBeg+1
                    iPosDatBeg=iPosDatBeg+1
                    sTmpLbl=sLine[iPosLblBeg:iPosLblEnd]
                    sTmpDat=sLine[iPosDatBeg:iPosDatEnd]
                    if self.iVerbose>5:
                        self.oLog.debug('  sLbl:%s sDat:%s'%(sTmpLbl,sTmpDat))
                    # ----- end:link info
                    # +++++ beg:
                    if iIsImage==0:
                        # +++++ beg:handle link
                        if self.iVerbose>(iVerboseLmt+5):
                            self.oLog.debug('  link found:%s'%(sTmpDat))
                        sReplDat=self.handleLink(sTmpDat,oGtrMD)
                        if sReplDat is not None:
                            sLine=sLine[:iPosDatBeg]+sReplDat+sLine[iPosDatEnd:]
                            iPosDatEnd=iPosDatEnd+(len(sTmpDat)-len(sReplDat))
                            #iPosBeg=iPosDatEnd+1
                        iPosBeg=iPosDatEnd+1
                        # ----- end:handle link
                    else:
                        # +++++ beg:handle link to image
                        if self.iVerbose>(iVerboseLmt+5):
                            self.oLog.debug('  image found:%s'%(sTmpDat))
                        oGtrMD.copyImage(self.sDN,sTmpDat)
                        iPosBeg=iPosDatEnd+1
                        # ----- end:handle link to image
                    # ----- end:
                # ----- end:process link
            # ----- end:check line
            iRet=1
        except:
            self.logTB()
        try:
            # +++++ beg:finalize
            iPosLF=sLine.rfind('\n')
            if self.iVerbose>10:
                self.oLog.debug('  sLine:>%s<'%(sLine[:iPosLF]))
            if self.iVerbose>0:
                self.oLog.debug('end:%s iRet:%d'%(sOrg,iRet))
            # ----- end:finalize
            return iRet,sLine
        except:
            self.logTB()
            return -1,sLine
    def prcExc(self,**kwargs):
        """processing execution
        ### parameter
            **kwargs ... flexible keyword argument
        ### return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            sOrg='ldmMD::prcExc'
            iVerboseLmt=0
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('beg:%s len(lMarker):%d'%(sOrg,
                                len(self.lMarker)))
            if self.iVerbose>0:
                self.oLog.debug('    kwargs:%r'%(kwargs))
            oGtrMD=kwargs.get('oGtrMD',None)
            # ----- end:initialize
            # +++++ beg:memorize file name
            self.sCurDN,self.sCurFN=os.path.split(self.sFN)
            if self.iVerbose>(iVerboseLmt+5):
                self.oLog.debug('    sCurDN:%r sCurFN:%r'%(self.sCurDN,self.sCurFN))
            if self.sCurDN is None:
                self.sCurDN='./'
            if len(self.sCurDN)==0:
                self.sCurDN='./'
            # ----- end:memorize file name
            # +++++ beg:processing end
            iRet=ldmStorageLine.prcExc(self,oGtrMD=oGtrMD)
            if self.iVerbose>(iVerboseLmt+5):
                self.oLog.debug('    %s iRet:%d iCnt:%d iAct:%d len(lLine):%d'%(sOrg,
                            iRet,self.iCnt,self.iAct,len(self.lLine)))
            ldmStorageLine.prcEnd(self)
            # +++++ beg:process referenced files
            lSubFN=self.getLstDef()
            if lSubFN is not None:
                for sSubFN in lSubFN:
                    self.addLine('\r\n')
                    # +++++ beg:get object to process referenced file
                    if oGtrMD is None:
                        iRetMD,oMD=self.getDat(sSubFN)
                    else:
                        iRetMD,oMD=oGtrMD.getDat(sSubFN)
                    if self.iVerbose>0:
                        self.oLog.debug('    %s oGtrMD:%r iRetMD:%d oMD:%r'%(sOrg,oGtrMD,iRetMD,oMD))
                    # ----- end:get object to process referenced file
                    # +++++ beg:process referenced file
                    if iRetMD>0:
                        # +++++ beg:process sub file
                        oMD.prcBeg(sSubFN,oDat=oMD,oRef=None)
                        iR=oMD.prcExc(oGtrMD=oGtrMD)
                        if iR>0:
                            iRet=iRet+1
                        oMD.prcEnd()
                        # ----- end:process sub file
                        # +++++ beg:add lines
                        lLines,iCnt=oMD.getLines()
                        for iAct,sL in lLines:
                            self.addLine(sL)
                        # ----- end:add lines
                    # ----- end:process referenced file
                #iR,sTmpFN,oTmpDat,oTmpRef=oMD.prcEnd()
            iR=self.clrDef()
            # ----- end:process referenced files
            if self.iVerbose>iVerboseLmt:
                self.oLog.debug('end:%s iRet:%d iCnt:%d'%(sOrg,
                                iRet,self.iCnt))
            # ----- end:processing end
            return iRet
        except:
            self.logTB()
            return -1



#----------------------------------------------------------------------------
# Name:         ldmWidCoreEvt.py
# Purpose:      ldmWidCoreEvt.py
#               core widget event
# Author:       Walter Obweger
#
# Created:      20200405
# CVS-ID:       $Id$
# Copyright:    (c) 2020 by Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import wx

gdSysEvtName={}
gTypeString=type('')

def ldmWidCoreKeyCode(evt):
    if wx.VERSION >= (2,8):
        return evt.GetKeyCode()
    else:
        return evt.KeyCode()

ldmEVT_WID_CORE_CMD=wx.NewEventType()
def EVT_WID_CORE_CMD(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_CMD,func)
def EVT_WID_CORE_CMD_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_CMD,func)
class ldmWidCoreCmd(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_CMD(<widget_name>, xxx)
    """
    def __init__(self,obj,cmd,data=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_CMD)
        self.cmd=cmd
        self.data=data
    def GetCmd(self):
        return self.cmd
    def GetData(self):
        return self.data
    GetDat=GetData

ldmEVT_WID_CORE_OK=wx.NewEventType()
def EVT_WID_CORE_OK(win,func):
    if hasattr(win,'GetWid'):
        win.__logDebug__(func)
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_OK,func)
def EVT_WID_CORE_OK_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_OK,func)
class ldmWidCoreOk(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_OK(<widget_name>, xxx)
    """
    def __init__(self,obj,res,data=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_OK)
        self.res=res
        self.data=data
    def GetResult(self):
        return self.res
    GetRlt=GetResult
    def GetData(self):
        return self.data
    GetDat=GetData

ldmEVT_WID_CORE_CANCEL=wx.NewEventType()
def EVT_WID_CORE_CANCEL(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_CANCEL,func)
def EVT_WID_CORE_CANCEL_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_CANCEL,func)
class ldmWidCoreCancel(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_CANCEL(<widget_name>, xxx)
    """
    def __init__(self,obj,res,data=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_CANCEL)
        self.res=res
        self.data=data
    def GetResult(self):
        return self.res
    GetRlt=GetResult
    def GetData(self):
        return self.data
    GetDat=GetData

ldmEVT_WID_CORE_FB=wx.NewEventType()
def EVT_WID_CORE_FB(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_FB,func)
def EVT_WID_CORE_FB_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_FB,func)
class ldmWidCoreFB(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_FB(<widget_name>, xxx)
    """
    def __init__(self,obj,fb,data=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_FB)
        self.fb=fb
        self.data=data
    def GetFB(self):
        return self.fb
    def GetData(self):
        return self.data
    GetDat=GetData

ldmEVT_WID_CORE_MOD=wx.NewEventType()
def EVT_WID_CORE_MOD(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_MOD,func)
def EVT_WID_CORE_MOD_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_MOD,func)
class ldmWidCoreMOD(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_MOD(<widget_name>, xxx)
    """
    def __init__(self,obj,res,data=None,gui=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_MOD)
        self.res=res
        self.data=data
        self.gui=gui
    def GetResult(self):
        return self.res
    GetRlt=GetResult
    def GetData(self):
        return self.data
    GetDat=GetData
    def GetGUI(self):
        return self.gui

ldmEVT_WID_CORE_NFY=wx.NewEventType()
def EVT_WID_CORE_NFY(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_NFY,func)
def EVT_WID_CORE_NFY_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_NFY,func)
class ldmWidCoreNFY(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_NFY(<widget_name>, xxx)
    """
    def __init__(self,obj,res,data=None,gui=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_NFY)
        self.res=res
        self.data=data
        self.gui=gui
    def GetResult(self):
        return self.res
    GetRlt=GetResult
    def GetData(self):
        return self.data
    GetDat=GetData
    def GetGUI(self):
        return self.gui

ldmEVT_WID_CORE_ADT=wx.NewEventType()
def EVT_WID_CORE_ADT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Connect(-1,-1,ldmEVT_WID_CORE_ADT,func)
def EVT_WID_CORE_ADT_DISCONNECT(win,func):
    if hasattr(win,'GetWid'):
        win=win.GetWid()
    win.Disconnect(-1,-1,ldmEVT_WID_CORE_ADT,func)
class ldmWidCoreADT(wx.PyEvent):
    """
    Posted Events:
        Tree Item selected event
            EVT_WID_CORE_ADT(<widget_name>, xxx)
    """
    def __init__(self,obj,res,old=None,new=None,gui=None):
        wx.PyEvent.__init__(self)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.SetEventType(ldmEVT_WID_CORE_ADT)
        self.res=res
        self.old=old
        self.new=new
        self.gui=gui
    def GetResult(self):
        return self.res
    GetRlt=GetResult
    def GetOld(self):
        return self.old
    def GetNew(self):
        return self.new
    def GetGUI(self):
        return self.gui

class ldmWidCoreEvt:
    _MAP_EVENT={}
    _MAP_EVENT_UNBIND={}
    def BindEvent(self,name,func,par=None):
        try:
            bDbg=self.GetVerboseDbg(100)
            if bDbg:
                self.logDbg('BindEvent name:%s',name)
            # ++++++++++
            # 20141125 wro: event mapping added
            evt=None
            if name in self._MAP_EVENT:
                evt=self._MAP_EVENT[name]
                if type(evt)==gTypeString:
                    name=evt
                    if name in self._MAP_EVENT:
                        evt=self._MAP_EVENT[name]
            else:
                #self.BindEventWid(par,name,func)
                #return
                evt=self.__get_widget_event__(name)
                if type(evt)==gTypeString:
                    name=evt
                    evt=self.__get_widget_event__(name)
                else:
                    wid=self.GetWid()
                    if par is None:
                        wid.Bind(evt,func)
                    else:
                        wid.Bind(evt,func,par)
                    return 1
                if evt is None:
                    return -2
            # ----------
            if evt is not None:
                if par is None:
                    evt(self,func)
                else:
                    evt(par,func)
                return 0
            else:
                self.logErr('evt name:%s not resolved',name)
        except:
            self.logTB()
        return -1
    def UnBindEvent(self,name,func,par=None):
        try:
            # ++++++++++
            # 20141125 wro: event mapping added
            evt=None
            if name in self._MAP_EVENT_UNBIND:
                evt=self._MAP_EVENT_UNBIND[name]
                if type(evt)==gTypeString:
                    name=evt
                    if name in self._MAP_EVENT_UNBIND:
                        evt=self._MAP_EVENT_UNBIND[name]
            # ----------
            if evt is not None:
                if par is None:
                    evt(self,func)
                else:
                    evt(par,func)
                return 0
            else:
                self.logErr('evt name:%s not resolved',name)
        except:
            self.logTB()
        return -1
    def __get_widget_event__(self,sK):
        try:
            bDbg=self.GetVerboseDbg(100)
            if bDbg:
                self.logDbg('__get_widget_event__ sK:%s',sK)
            if sK=='txt':
                return wx.EVT_TEXT
            elif sK=='ent':
                return wx.EVT_TEXT_ENTER
            elif sK=='btn':
                return wx.EVT_BUTTON
            elif sK=='mn':
                return wx.EVT_MENU
            elif sK=='tg':
                return wx.EVT_TOGGLEBUTTON
            elif sK=='choice':
                return wx.EVT_CHOICE
            elif sK=='check':
                return wx.EVT_CHECKBOX
            elif sK=='checkBox':
                return wx.EVT_CHECKLISTBOX
            elif sK=='lstSel':
                return wx.EVT_LISTBOX
            elif sK=='lstDblclk':
                return wx.EVT_LISTBOX_DCLICK
            elif sK=='lstCtrlSel':
                return wx.EVT_LIST_ITEM_SELECTED
            elif sK=='lstCtrlDesel':
                return wx.EVT_LIST_ITEM_DESELECTED
            elif sK=='lstCtrlDblClk':
                return wx.EVT_LIST_ITEM_ACTIVATED
            elif sK=='lstCtrlAct':
                return wx.EVT_LIST_ITEM_ACTIVATED
            elif sK=='lstCtrlActivate':
                return wx.EVT_LIST_ITEM_ACTIVATED
            elif sK=='lstCtrlColClk':
                return wx.EVT_LIST_COL_CLICK
            elif sK=='lstCtrlColClkLf':
                return wx.EVT_LIST_COL_CLICK
            elif sK=='lstCtrlColClkRg':
                return wx.EVT_LIST_COL_RIGHT_CLICK
            elif sK=='lstCtrlKey':
                return wx.EVT_LIST_KEY_DOWN
            elif sK=='trCtrlSel':
                return wx.EVT_TREE_ITEM_SELECTED
            elif sK=='trCtrlDesel':
                return wx.EVT_TREE_ITEM_DESELECTED
            elif sK=='trCtrlDbClk':
                return wx.EVT_TREE_ITEM_ACTIVATED
            elif sK=='trCtrlAct':
                return wx.EVT_TREE_ITEM_ACTIVATED
            elif sK=='trCtrlActivate':
                return wx.EVT_TREE_ITEM_ACTIVATED
            elif sK=='trCtrlDel':
                return wx.EVT_TREE_DELETE_ITEM
            elif sK=='trCtrlCollapsing':
                return wx.EVT_TREE_ITEM_COLLAPSING
            elif sK=='trCtrlCollapsed':
                return wx.EVT_TREE_ITEM_COLLAPSED
            elif sK=='trCtrlExpanding':
                return wx.EVT_TREE_ITEM_EXPANDING
            elif sK=='trCtrlExpanded':
                return wx.EVT_TREE_ITEM_EXPANDED
            elif sK=='trCtrlSelChg':
                return wx.EVT_TREE_SEL_CHANGING
            elif sK=='trCtrlSel':
                return wx.EVT_TREE_SEL_CHANGED
            elif sK=='trCtrlLclk':
                return wx.EVT_TREE_ITEM_LEFT_CLICK
            elif sK=='trCtrlMclk':
                return wx.EVT_TREE_ITEM_MIDDLE_CLICK
            elif sK=='trCtrlRclk':
                return wx.EVT_TREE_ITEM_RIGHT_CLICK
            elif sK=='trCtrlMenu':
                return wx.EVT_TREE_ITEM_MENU
            elif sK=='trCtrlDragBeg':
                return wx.EVT_TREE_BEGIN_DRAG
            elif sK=='trCtrlDragBegRg':
                return wx.EVT_TREE_BEGIN_RDRAG
            elif sK=='trCtrlDragEnd':
                return wx.EVT_TREE_END_DRAG
            elif sK=='trCtrlKey':
                return wx.EVT_TREE_KEY_DOWN
            elif sK=='char':
                return wx.EVT_CHAR
            elif sK=='char_hook':
                return wx.EVT_CHAR_HOOK
            elif sK=='keyDn':
                return wx.EVT_KEY_DOWN
            elif sK=='keyUp':
                return wx.EVT_KEY_UP
            elif sK=='mouse':
                return wx.EVT_MOUSE_EVENTS
            elif sK=='wheel':
                return wx.EVT_MOUSEWHEEL
            elif sK=='rgDbl':
                return wx.EVT_RIGHT_DCLICK
            elif sK=='mdDbl':
                return wx.EVT_MIDDLE_DCLICK
            elif sK=='lfDbl':
                return wx.EVT_LEFT_DCLICK
            elif sK=='motion':
                return wx.EVT_MOTION
            elif sK=='rgDn':
                return wx.EVT_RIGHT_DOWN
            elif sK=='rgUp':
                return wx.EVT_RIGHT_UP
            elif sK=='mdUp':
                return wx.EVT_MIDDLE_UP
            elif sK=='mdDn':
                return wx.EVT_MIDDLE_DOWN
            elif sK=='lfUp':
                return wx.EVT_LEFT_UP
            elif sK=='lfDn':
                return wx.EVT_LEFT_DOWN
            elif sK=='enter':
                return wx.EVT_ENTER_WINDOW
            elif sK=='leave':
                return wx.EVT_LEAVE_WINDOW
            elif sK=='move':
                return wx.EVT_MOVE
            elif sK=='size':
                return wx.EVT_SIZE
            elif sK=='erase':
                return wx.EVT_ERASE_BACKGROUND
            elif sK=='paint':
                return wx.EVT_PAINT
            elif sK=='focusKill':
                return wx.EVT_KILL_FOCUS
            elif sK=='focusSet':
                return wx.EVT_SET_FOCUS
            elif sK=='help':
                return wx.EVT_HELP
            elif sK=='colorChanged':
                return wx.EVT_SYS_COLOUR_CHANGED
            # dialog events
            elif sK=='idle':
                return wx.EVT_IDLE
            elif sK=='navKey':
                return wx.EVT_NAVIGATION_KEY
            elif sK=='iconize':
                return wx.EVT_ICONIZE
            elif sK=='maximize':
                return wx.EVT_MAXIMIZE
            elif sK=='dropFiles':
                return wx.EVT_DROP_FILES
            elif sK=='close':
                return wx.EVT_CLOSE
            elif sK=='activate':
                return wx.EVT_ACTIVATE
            elif sK=='init':
                return wx.EVT_INIT_DIALOG
            elif sK=='ctxmn':
                return wx.EVT_CONTEXT_MENU
            elif sK=='cxtmn':
                return wx.EVT_CONTEXT_MENU
            return None
        except:
            self.logTB()
        return None
    def GetSysEvtNameByEvt(self,evt):
        try:
            global gdSysEvtName
            bDbg=self.GetVerboseDbg(100)
            if bDbg:
                self.LogDbg('GetSysEvtNameByEvt')
            iEvtId=evt.GetEventType()
            if iEvtId in gdSysEvtName:
                return gdSysEvtName[iEvtId]
            for s in dir(wx):
                if s.startswith('EVT'):
                    #print s
                    e=getattr(wx,s)
                    if hasattr(e,'evtType'):
                        lEvtId=e.evtType
                        #print iEvtId,lEvtId
                        if iEvtId in lEvtId:
                            gdSysEvtName[iEvtId]=s[:]
                            return s
        except:
            self.logTB()
    def BindEventDict(self,wid,dEvt):
        try:
            bDbg=self.GetVerboseDbg(100)
            if bDbg:
                self.logDbg('BindEventDict wid:%r dEvt:%r',wid,dEvt)
            for sK,fct in dEvt.iteritems():
                if wid is None:
                    self.BindEvent(sK,fct)
                else:
                    self.BindEventWid(wid,sK,fct)
        except:
            self.logTB()
    def BindEventWid(self,wid,sEvt,fct):
        try:
            bDbg=self.GetVerboseDbg(100)
            if bDbg:
                self.logDbg('BindEventWid sEvt:%s fct:%r wid:%r',sEvt,fct,wid)
            evt=self.__get_widget_event__(sEvt)
            if type(evt)==gTypeString:
                sEvt=evt
                evt=self.__get_widget_event__(sEvt)
            if evt is not None:
                t=type(wid)
                if t==gTypeString:
                    w=self.__dict__.get(wid,None)
                elif t==types.UnicodeType:
                    w=self.__dict__.get(wid,None)
                else:
                    w=wid
                if w is None:
                    return 
                w.Bind(evt,fct)
        except:
            self.logTB()




#----------------------------------------------------------------------------
# Name:         lmdOS.py
# Purpose:      OS class
#
# Author:       Walter Obweger
#
# Created:      20200322
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import sys
import os
import os.path
import time
import stat
import traceback

def newSha():
    """python 3 change fix
    """
    if sys.version_info  >= (2,6,0):
        import hashlib
        return hashlib.sha1()
    else:
        import sha
        return sha.new()

def getSha(sFN,iMB=-1,iBlk=4096,sDN=None,oLog=None,iVerbose=0):
    """get SHA limited
    ### parameter
        sFN     ... file name
        iMB     ... max size to read
                    + <=0 all data
                    + >0 up to maximum in MB
        iBlk    ... block size
        oLog    ... logging object
    ### return
        >0  ... okay content read
        =0  ... okay no data
        <0  ... error
    """
    try:
        if sDN is not None:
            sFullFN=os.path.join(sDN,sFN)
        else:
            sFullFN=sFN
        if oLog is None:
            iVerbose=0
        with open(sFullFN,'rb') as oFile:
            if iVerbose>0:
                oLog.debug('sFullFN:%s'%(sFullFN))
            # +++++ beg:initialize
            oStat=os.stat(sFullFN)
            oSha=newSha()
            if iBlk<64:
                iBlk=64
            if iMB==0:
                iSz=0
            elif iMB<=0:
                iSz=oStat.st_size
                iBlkSz=4096
            else:
                iSz=iMB<<20
                if iSz>oStat.st_size:
                    iSz=oStat.st_size
            if iSz==0:
                iEnd=2
            else:
                iEnd=0
            # ----- end:initialize
            # +++++ beg:read file data
            iPrv=0
            iCur=0
            while iEnd==0:
                try:
                    iPrv=iCur
                    if iVerbose>0:
                        oLog.debug('iBlk:%d iPrv:%d'%(iBlk,iPrv))
                    sDat=oFile.read(iBlk)
                    iCur=oFile.tell()
                    iRd=iCur-iPrv
                    if iVerbose>0:
                        oLog.debug('iRd:%d iCur:%d'%(iRd,iCur))
                    if iRd>0:
                        oSha.update(sDat)
                        if iRd<iBlk:
                            iEnd=1
                        if iCur>iSz:
                            iEnd=3
                    else:
                        iEnd=2
                except:
                    if oLog is not None:
                        oLog.error(traceback.format_exc())
                    iEnd=-1
            # ----- end:read file data
            # +++++ beg:deliver fingerprint value
            if iEnd==1:         # regular
                return oSha.hexdigest()
            elif iEnd==2:       # empty
                return "---"
            elif iEnd==3:       # limit reached
                return oSha.hexdigest()
            elif iEnd==-1:      # exception
                return '!!!'
            return '???'        # you're not supposed to be here
            # ----- end:deliver fingerprint value
    except:
        if oLog is not None:
            oLog.error(traceback.format_exc())
        return '???!!!'

def getNow(iUTC=1,sFmt='%Y%m%d_%H%M%S'):
    """get string current time
    ### parameter
        iUTC    ... in universal time cooradinated
        sFmt    ... format
    ### return
        str ... string representing current time
    """
    if iUTC>0:
        zNow=time.gmtime(time.time())
    else:
        zNow=time.localtime(time.time())
    sNow=time.strftime(sFmt,zNow)
    return sNow

def getNiceFN(sFN,iKind=0,oLog=None):
    """build file name and create required directories
    ### parameter
        sFN     ... file name
    ### return
        >0  ... okay content read
        =0  ... okay no data
        <0  ... error
    """
    try:
        # +++++ beg:initialize
        iRet=0
        iLen=len(sFN)
        lRet=[]
        # ----- end:initialize
        for s in sFN:
            iFound=0
            iOrd=ord(s)
            if (iOrd>=65) and (iOrd<=90):
                iFound=1
            elif (iOrd>=97) and (iOrd<=122):
                iFound=1
            elif (iOrd>=48) and (iOrd<=57):
                iFound=1
            elif s in ['.',',','_','-']:
                iFound=1
            else:
                pass
            if iFound>0:
                lRet.append(s)
            else:
                if iKind==0:
                    lRet.append('_')
        return ''.join(lRet)
    except:
        if oLog is not None:
            oLog.error(traceback.format_exc())
        return -1,None
    
def bldFN(sFN,sDN=None,sExt='json',sSfx=None,oLog=None,iVerbose=0):
    """build file name and create required directories
    ### parameter
        sFN     ... file name
        sDN     ... directory name
        sExt    ... file name extension
        sSfx    ... suffix
        oLog    ... logging object
        iVerbose .. verbose detail level
    ### return
        >0  ... okay content read
        =0  ... okay no data
        <0  ... error
    """
    try:
        # +++++ beg:initialize
        iRet=0
        iVrb=0
        if oLog is not None:
            if iVerbose>0:
                sOrg='ldmOS::bldFN'
                oLog.debug('beg:%s iVerbose:%d'%(sOrg,iVerbose))
                iVrb=1
        # ----- end:initialize
        # +++++ beg:ensure file name
        if sFN is None:
            sFN=getNow()
        elif sFN==0:
            sFN=getNow(iUTC=0)
        elif sFN==1:
            sFN=getNow(iUTC=1)
        # ----- end:ensure file name
        # +++++ beg:build file name
        if sDN is not None:
            sFullFN=os.path.join(sDN,sFN)
        else:
            sFullFN=sFN
        if sExt is not None:
            iLenExt=len(sExt)+1
            if sFullFN.endswith(sExt):
                lFN=[sFullFN[:-iLenExt]]
            else:
                lFN=[sFullFN]
        else:
            lFN=[sFullFN]
        if sSfx is not None:
            sSfxNice=getNiceFN(sSfx)
            lFN.append(sSfxNice)
        if sExt is not None:
            lFN.append(sExt)
        sFullFN='.'.join(lFN)
        if iVrb>0:
            oLog.debug('   :sFullFN:%s'%(sFullFN))
        # ----- end:build file name
        # +++++ beg:create folder
        try:
            sTmpDN,sTmpFN=os.path.split(sFullFN)
            if len(sTmpDN)>0:
                if os.path.exists(sTmpDN)==False:
                    os.makedirs(sTmpDN)
            iRet=1    
        except:
            if oLog is not None:
                oLog.error(traceback.format_exc())
        # ----- end:create folder
        # +++++ beg:finalize
        if iVrb>0:
            oLog.debug('end:%s iRet:%d'%(sOrg,iRet))
        # ----- end:finalize
        return iRet,sFullFN
    except:
        if oLog is not None:
            oLog.error(traceback.format_exc())
        return -1,None



#----------------------------------------------------------------------------
# Name:         ldmWidSizeRspWid.py
# Purpose:      ldmWidSizeRspWid.py
#               GUI widget respond on size change
# Author:       Walter Obweger
#
# Created:      20200405
# CVS-ID:       $Id$
# Copyright:    (c) 2020 by Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import wx

class ldmWidSizeRspWid(wx.Control):
    def __init__(self, parent, id=-1,pos=None,size=None, mgr=None):
        wx.Control.__init__(self, parent, id,
                        pos or wx.DefaultPosition,
                        size or wx.DefaultSize,
                        style=wx.NO_BORDER)
        self.mgrGui = mgr

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_SIZE, self.OnSize)
    def OnPaint(self, event):
        dc = wx.PaintDC(self)
        size = self.GetClientSize()


        dc.SetFont(wx.NORMAL_FONT)
        dc.SetBrush(wx.WHITE_BRUSH)
        dc.SetPen(wx.WHITE_PEN)
        
        iWdh=size.x
        iHgt=size.y
        iCtrX=iWdh/2
        iCtrY=iHgt/2
        iWdh-=1
        iHgt-=1
        
        dc.DrawRectangle(0, 0, iWdh, iHgt)
        dc.SetPen(wx.BLACK_PEN)
        #dc.DrawLine(0, 0, size.x, size.y)
        #dc.DrawLine(0, size.y, size.x, 0)
        #dc.SetPen(wx.LIGHT_GREY_PEN)
        iMrk=20
        dc.DrawPolygon([(0,0),(iMrk,0),(0,iMrk)])
        dc.DrawPolygon([(iWdh,0),(iWdh-iMrk,0),(iWdh,iMrk)])
        dc.DrawPolygon([(iWdh,iHgt),(iWdh-iMrk,iHgt),(iWdh,iHgt-iMrk)])
        dc.DrawPolygon([(0,iHgt),(iMrk,iHgt),(0,iHgt-iMrk)])
        
        dc.SetPen(wx.LIGHT_GREY_PEN)
        dc.DrawPolygon([
            (iMrk,iMrk),
            (iWdh-iMrk,iMrk),
            (iWdh-iMrk,iHgt-iMrk),
            (iMrk,iHgt-iMrk),
            ])
        dc.SetPen(wx.BLACK_PEN)
        lStr=[]
        def addStr(l,s,iHghGap=3):
            iWdhTxt, iHgtTxt = dc.GetTextExtent(s)
            iHgtTxt += iHghGap
            l.append([s,iWdhTxt, iHgtTxt])
        addStr(lStr,"Size: %d x %d"%(size.x, size.y))
        
        if self.mgrGui:
            pi = self.mgrGui.GetPane(self)
            addStr(lStr,"layer: %d"%(pi.dock_layer))
            addStr(lStr,"dock: %d row: %d pos: %d"%(pi.dock_direction,pi.dock_row,pi.dock_pos))
            addStr(lStr,"prop: %d"%(pi.dock_proportion))
        iHgtTxt=0
        for it in lStr:
            iHgtTxt+=it[2]
        iCtrY-=iHgtTxt>>1
        iHgtTxt=0
        for s,iWdhTxt,iHgtTxt in lStr:
            dc.DrawText(s, iCtrX - (iWdhTxt/2), iCtrY)
            iCtrY+=iHgtTxt
    def OnEraseBackground(self, event):
        pass
    def OnSize(self, event):
        self.Refresh()



#----------------------------------------------------------------------------
# Name:         lndStorage.py
# Purpose:      storage base class
#
# Author:       Walter Obweger
#
# Created:      20200104
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import logging
import traceback

from lindworm.logUtil import logGet

class ldmStorage:
    def __init__(self,sLogger='',iLevel=logging.DEBUG,iVerbose=0):
        """constructor
        """
        self.iVerbose=iVerbose
        self.__initDat__()
        self.oLog=logGet(sLogger,iLevel)
    def __initDat__(self):
        """initialize data properties
        """
        self.lFN=[]         # list files been processed
        self.dFN={}         # dictionary processed files, performance
    def clr(self):
        self.__initDat__()
    def prcIsDone(self,sFN):
        """check file name has been processed already  
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        if sFN in self.dFN:
            iRet=1
        else:
            iRet=0
        self.oLog.debug('%s sFN:%s iRet:%d'%('ldmStorage::prcIsDone',
                        sFN,iRet))
        return iRet
    def _tpl(self,iVerbose=-1):
        """
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            logging.debug('beg: iVerbose:%d'%(iVerbose))
            iRet=0
            # ----- end:
            # +++++ beg:
            # ----- end:
            # +++++ beg:
            logging.debug('end:  iRet:%d'%(iRet))
            # ----- end:
            return iRet
        except:
            logging.error(traceback.format_exc())
            return -1
    def _tpl2(self):
        """
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            if self.iVerbose>0:
                self.oLog.debug('beg: iVerbose:%d'%(self.iVerbose))
            iRet=0
            # ----- end:
            # +++++ beg:
            # ----- end:
            # +++++ beg:
            if self.iVerbose>0:
                self.oLog.debug('end:  iRet:%d'%(iRet))
            # ----- end:
            return iRet
        except:
            self.oLog.error(traceback.format_exc())
            return -1



#----------------------------------------------------------------------------
# Name:         lndMD.py
# Purpose:      MD class
#
# Author:       Walter Obweger
#
# Created:      20200108
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import logging
import traceback

class ldmStr:
    def findBegEnd(self,sLine,iLen,iPos,cBeg='[',cEnd=']',iVerbose=-1):
        """find whitespace begin or end
        sLine   ... string to search
        iLen    ... string length
        iPos    ... position to start search
        iMode   ... search mode, 1= find begin , 0+ find end 
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            if iVerbose>0:
                logging.debug('beg:%s iVerbose:%d'%('ldmStr::findWhiteSpace',iVerbose))
            iRet=-1
            iBeg=-1
            iEnd=-1
            lWhiteSpace=[' ']
            # ----- end:initialize
            # +++++ beg:
            iBeg=sLine.find(cBeg,iPos)
            if iBeg>=0:
                iEnd=sLine.find(cEnd,iBeg+1)
                if iEnd>0:
                    return iBeg,iEnd
            # ----- end:
            # +++++ beg:finalize
            if iVerbose>0:
                logging.debug('end:%s  iRet:%d iBeg:%d iEnd:%d'%('ldmStr::findWhiteSpace',iRet,
                                                    iBeg,iEnd))
            # ----- end:finalize
            return iBeg,iEnd
        except:
            logging.error(traceback.format_exc())
            return -1,-1
    def findWhiteSpace(self,sLine,iLen,iPos,iMode=1,iVerbose=-1):
        """find whitespace begin or end
        sLine   ... string to search
        iLen    ... string length
        iPos    ... position to start search
        iMode   ... search mode, 1= find begin , 0+ find end 
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:
            if iVerbose>0:
                logging.debug('beg:%s iVerbose:%d'%('ldmStr::findWhiteSpace',iVerbose))
            iRet=-1
            lWhiteSpace=[' ']
            # ----- end:
            # +++++ beg:
            if iLen<0:
                iLen=len(sLine)
            if iPos<0:
                iPos=0
            iOfs=iPos
            iFound=0
            while (iFound==0) and (iOfs<iLen):
                sChk=sLine[iOfs]
                if iMode>0:
                    if sChk not in lWhiteSpace:
                        iOfs=iOfs+1
                    else:
                        iFound=1
                        iRet=iOfs
                else:
                    if sChk in lWhiteSpace:
                        iOfs=iOfs+1
                    else:
                        iFound=1
                        iRet=iOfs
            if iFound==0:
                if iOfs>=iLen:
                    iRet=sLine.rfind('\r')
                    if iRet<0:
                        iRet=sLine.rfind('\n')
                        if iRet<0:
                            iRet=iLen
            # ----- end:
            # +++++ beg:
            if iVerbose>0:
                logging.debug('end:%s  iRet:%d'%('ldmStr::findWhiteSpace',iRet))
            # ----- end:
            return iRet
        except:
            logging.error(traceback.format_exc())
            return -1
    def replaceIgnore(self,sNav,sReplIgnore=''):
        """replace strange characters to ensure proper 
        header navigation.
        return
            >0  ... okay processing done
            =0  ... okay nop
            <0  ... error
        """
        try:
            # +++++ beg:initialize
            iLen=len(sNav)
            lNav=[]
            # ----- end:initialize
            # +++++ beg:loop through string
            for iOfs in range(0,iLen):
                cTmp=sNav[iOfs]
                iTmp=ord(cTmp)
                if (iTmp>=65) and (iTmp<=90):
                    pass
                elif (iTmp>=97) and (iTmp<=122):
                    pass
                elif (iTmp>=48) and (iTmp<=57):
                    pass
                elif cTmp in ['.']:
                    pass
                else:
                    cTmp=sReplIgnore
                lNav.append(cTmp)
            # ----- end:loop through string
            # +++++ beg:build result
            sNav=''.join(lNav)
            # ----- end:build result
            return sNav
        except:
            logging.error(traceback.format_exc())
            return ''



#----------------------------------------------------------------------------
# Name:         ldmWidPanel.py
# Purpose:      ldmWidPanel.py
#               GUI widget respond on size change
# Author:       Walter Obweger
#
# Created:      20200413
# CVS-ID:       $Id$
# Copyright:    (c) 2020 by Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import wx
from lindworm.ldmWidCore import ldmWidCore
import lindworm.ldmGui as ldmGui

class ldmWidPanel(ldmWidCore):
    def getEvtDat(self,evt):
        try:
            # +++++ beg:
            iRet=0
            # ----- end:
            o=evt.GetEventObject()
            iId=evt.GetId()
            self.logDbg('getEvtWid iId:%d 0x%x',iId,iId)
            oDat=self.__getRegDat__(iId)
            return oDat
            w=o.FindTool(iId)
            return w
        except:
            self.logTB()
            return None
    def getFileOpen(self):
        try:
            # +++++ beg:
            iRet=0
            # ----- end:
            return iRet
        except:
            self.logTB()
            return -1
    def prcCmd(self,sCmd):
        try:
            # +++++ beg:
            self.logDbg('ldmWidPanel::prcCmd sCmd:%r',sCmd)
            # ----- end:
            return 0
        except:
            self.logTB()
            return -1
    def OnCmd(self,evt):
        evt.Skip()
        try:
            # +++++ beg:initialize
            oDat=self.getEvtDat(evt)
            self.logDbg('ldmWidPanel::OnCmd oDat:%r',oDat)
            iRet=self.prcCmd(oDat)
            self.logDbg('ldmWidPanel::OnCmd prcCmd iRet:%r',iRet)
            # ----- end:initialize
        except:
            self.logTB()
    def OnBwsDN(self,evt):
        evt.Skip()
        try:
            # +++++ beg:initialize
            oDat=self.getEvtDat(evt)
            self.logDbg('ldmWidPanel::OnBwsDN oDat:%r',oDat)
            sWidName=oDat[0]
            sKind=oDat[1]
            # ----- end:initialize
            # +++++ beg:get widget and value
            wSub=getattr(self,sWidName)
            sOldDN=wSub.GetValue()
            # ----- end:get widget and value
            # +++++ beg:get DN
            iRet,sNewDN=ldmGui.getDN(sOldDN,self.GetWid(),
                        'choose source directory')
            # ----- end:get DN
            # +++++ beg:update data
            if iRet>0:
                wSub.SetValue(sNewDN)
                self.logDbg('fin:OnBwsDN iRet:%d %s:%s',
                            iRet,
                            sWidName,sNewDN)
            # ----- end:update data
        except:
            self.logTB()
    def OnBwsFN(self,evt):
        evt.Skip()
        try:
            # +++++ beg:initialize
            oDat=self.getEvtDat(evt)
            self.logDbg('ldmWidPanel::OnBwsFN oDat:%r',oDat)
            iLenDat=len(oDat)
            sWidName=oDat[0]
            sKind=oDat[1]
            sWidLnk=None
            # ----- end:initialize
            # +++++ beg:get widget and value
            wSub=getattr(self,sWidName)
            sOldFN=wSub.GetValue()
            # ----- end:get widget and value
            # +++++ beg:get FN
            WILDCARD_JSON=[
                    "All files (*.*)|*.*",
                    "json file (*.json)|*.json",
                    "text file (*.txt)|*.txt",
                    "log file (*.log)|*.log",
                    "Excel file (*.xlsx)|*.xlsx",
                    "Excel file obsolete (*.xls)|*.xls",
                    ]

            iRet,sNewFN=ldmGui.getFN(sOldFN,self.GetWid(),
                        'choose file name',
                        lWildCard=WILDCARD_JSON)
            # ----- end:get FN
            # +++++ beg:update data
            if iRet>0:
                if iLenDat>2:
                    sWidLnk=oDat[2]
                if sWidLnk is not None:
                    # +++++ beg:handle linked widget, directory part
                    wLnk=getattr(self,sWidLnk)
                    sTmpDN,sTmpFN=ldmGui.getSplitFN(sNewFN)
                    if (sTmpDN is not None) and (sTmpFN is not None):
                        wLnk.SetValue(sTmpDN)
                        wSub.SetValue(sTmpFN)
                        self.log('fin:OnBwsFN iRet:%d %s:%s %s:%s',
                                    iRet,
                                    sWidLnk,sTmpDN,
                                    sWidName,sTmpFN)
                    else:
                        self.logErr('sTmpDN:%r sTmpFN:%r',sTmpDN,sTmpFN)
                        wLnk.SetValue('')
                        wSub.SetValue(sNewFN)
                        self.logDbg('fin:OnBwsFN iRet:%d %s:%s',
                                    iRet,sWidName,sNewFN)
                    # ----- end:handle linked widget, directory part
                else:
                    wSub.SetValue(sNewFN)
                    self.logDbg('fin:OnBwsFN iRet:%d %s:%s',
                                iRet,sWidName,sNewFN)
            # ----- end:update data
        except:
            self.logTB()
    def __getRegDat__(self,iId):
        try:
            # +++++ beg:
            if iId in self.dIdDat:
                oDat=self.dIdDat[iId]
            else:
                oDat=None
            # ----- end:
            return oDat
        except:
            self.logTB()
            return None
    def __setRegDat__(self,oDat):
        try:
            # +++++ beg:
            self.logDbg('iId:%d oDat:%r',self.iId,oDat)
            self.dIdDat[self.iId]=oDat
            self.iId+=1
            # ----- end:
        except:
            self.logTB()
    def __bldWid__(self,tDef):
        try:
            # +++++ beg:
            iExp=0
            iLen=len(tDef)
            sCls=tDef[0]
            if iLen>1:
                sName=tDef[1]
            else:
                sName=None
            if iLen>2:
                sVal=tDef[2]
            else:
                sVal=None
            if iLen>3:
                sLnk=tDef[3]
            else:
                sLnk=None
            self.logDbg('__bldWid__ tDef:%r',tDef)
            if sVal is None:
                sVal=''
            # ----- end:
            # +++++ beg:
            wPar=self.GetWid()
            if sCls=='lbl':
                wSub=wx.StaticText(wPar, wx.ID_ANY, sVal)
            elif sCls=='gag':
                try:
                    iMax=int(sVal)
                except:
                    iMax=1000
                wSub=wx.Gauge(wPar, wx.ID_ANY, range=iMax,
                            size=(-1,8))
                iExp=1
            elif sCls=='txt':
                wSub=wx.TextCtrl(wPar, wx.ID_ANY, sVal)
                #self.Bind(wx.EVT_TEXT_ENTER, self.OnSrcDnEnter, self.txtSrcDN)
                iExp=1
            elif sCls=='txtLn':
                wSub=wx.TextCtrl(wPar, wx.ID_ANY, sVal,style=wx.TE_MULTILINE)
                iExp=1
            elif sCls=='txtRd':
                wSub=wx.TextCtrl(wPar, wx.ID_ANY, sVal,style=wx.TE_READONLY)
                iExp=1
            elif sCls=='spn':
                lLmt=sVal[1]
                wSub=wx.SpinCtrl(wPar, self.iId, sVal[0],
                                min=int(lLmt[0]), max=int(lLmt[-1]))
                self.__setRegDat__(sVal)
                iExp=1
            elif sCls=='lst':
                iStyle=wx.LC_HRULES | wx.LC_REPORT | wx.LC_VRULES
                wSub=wx.ListCtrl(wPar, self.iId,size=(80,40),style=iStyle)
                self.__setRegDat__(sVal)
                iExp=1
            elif sCls=='chc':
                lChc=sVal[1]
                wSub=wx.Choice(wPar, self.iId,choices=lChc)
                try:
                    iIdx=lChc.index(sVal[0])
                    wSub.SetSelection(iIdx)
                except:
                    wSub.SetSelection(0)
                self.__setRegDat__(sVal)
                iExp=1
            elif sCls=='cb':
                wSub=wx.Button(wPar, self.iId, sVal)
                self.__setRegDat__(sVal)
                wPar.Bind(wx.EVT_BUTTON, self.OnCmd, wSub)
            elif sCls=='cbBmp':
                wSub=wx.BitmapButton(wPar, self.iId, sVal)
                self.__setRegDat__(sLnk)
                wPar.Bind(wx.EVT_BUTTON, self.OnCmd, wSub)
            elif sCls=='cbd':       # directory
                wSub=wx.Button(self.GetWid(), self.iId, '...')
                self.__setRegDat__((sVal,'browse_dir'))
                wPar.Bind(wx.EVT_BUTTON, self.OnBwsDN, wSub)
            elif sCls=='cbf':       # file open
                wSub=wx.Button(wPar, self.iId, '...')
                self.__setRegDat__((sVal,'browse_file_open',sLnk))
                wPar.Bind(wx.EVT_BUTTON, self.OnBwsFN, wSub)
            else:
                wSub=None
            # ----- end:
            # +++++ beg:
            if wSub is not None:
                if sName is not None:
                    if getattr(self,sName,None) is None:
                        setattr(self,sName,wSub)
                    else:
                        self.logErr('sName:%s already used'%(sName))
            # ----- end:
            return wSub,iExp
        except:
            self.logTB()
            return None,0
    def __initCls__(self,**kwargs):
        self.clsWid=wx.Panel
    def __initWid__(self,**kwargs):
        try:
            # +++++ beg:
            self.iId=1000
            self.dIdDat={}
            self.oSzMain=None
            self.logDbg('__initWid__')
            style=wx.TAB_TRAVERSAL
            _args,_kwargs=self.GetWidArgs(kwargs,
                        ['id','name','parent','pos','size','style'],
                        {'pos':(0,0),'size':(-1,-1),'style':style})
            self.wid=self.clsWid(*_args,**_kwargs)
            self.__initSizer__(**kwargs)
            # ----- end:
            # +++++ beg:
            #self.logDbg('check lWid')
            lWid=kwargs.get('lWid',None)
            if lWid is not None:
                for tDef in lWid:
                    #self.logDbg('build')
                    #self.logDbg('build %r',tDef)
                    wSub,iExp=self.__bldWid__(tDef)
                    iR=self.__addSizerWid__(tDef,wSub,iExp)
            # ----- end:
        except:
            self.logTB()
    def __initSizer__(self,**kwargs):
        try:
            bDbg=self.GetVerboseDbg(20)
            if bDbg:
                self.logDbg('__initLayout__')
            self.oSzMain=wx.BoxSizer(wx.HORIZONTAL)
        except:
            self.logTB()
    def __addSizerWid__(self,tDef,wSub,iExp):
        try:
            iRet=0
            if wSub is None:
                self.oSzMain.AddSpacer(5)
            else:
                if iExp<=0:
                    self.oSzMain.Add(wSub,0,0,0)
                else:
                    self.oSzMain.Add(wSub,iExp,wx.EXPAND,0)
        except:
            self.logTB()
    def __initLayout__(self,**kwargs):
        try:
            bDbg=self.GetVerboseDbg(20)
            if bDbg:
                self.logDbg('__initLayout__')

            if self.oSzMain is not None:
                self.wid.SetSizer(self.oSzMain)
                self.wid.Layout()
        except:
            self.logTB()

class ldmWidPanelVert(ldmWidPanel):
    def __initSizer__(self,**kwargs):
        try:
            bDbg=self.GetVerboseDbg(20)
            if bDbg:
                self.logDbg('__initLayout__')
            self.oSzMain=wx.BoxSizer(wx.VERTICAL)
        except:
            self.logTB()

class ldmWidPanelFlxGrd(ldmWidPanel):
    def __initSizer__(self,**kwargs):
        try:
            bDbg=self.GetVerboseDbg(20)
            if bDbg:
                self.logDbg('__initLayout__')
            iRow=kwargs.get('iRow',None)
            iCol=kwargs.get('iCol',2)

            if iRow is not None:
                self.oSzMain=wx.FlexGridSizer(iRow, iCol, 0, 0)
            else:
                self.oSzMain=wx.FlexGridSizer(iCol, 0, 0)
        except:
            self.logTB()
    def __initLayout__(self,**kwargs):
        try:
            bDbg=self.GetVerboseDbg(20)
            if bDbg:
                self.logDbg('__initLayout__')
            if self.oSzMain is not None:
                lCol=kwargs.get('lCol',None)
                if lCol is not None:
                    for iCol in lCol:
                        self.oSzMain.AddGrowableCol(iCol)
                lRow=kwargs.get('lRow',None)
                if lRow is not None:
                    for iRow in lRow:
                        self.oSzMain.AddGrowableRow(iRow)
                self.wid.SetSizer(self.oSzMain)
                self.wid.Layout()
        except:
            self.logTB()




#----------------------------------------------------------------------------
# Name:         lmdGuiNty.py
# Purpose:      gui notify class
#
# Author:       Walter Obweger
#
# Created:      20200404
# CVS-ID:       $Id$
# Copyright:    Walter Obweger
# Licence:      MIT
#----------------------------------------------------------------------------

import time
import wx

from six.moves import _thread as sixThd

from lindworm.logUtil import ldmUtilLog

class ldmGuiNty:
    def __init__(self,iVal=0,iMin=0,iMax=100):
        self.sPhase=''
        self.sStatus=''
        self.iVal=iVal
        self.iMin=iMin
        self.iMax=iMax
        self.iSchedule=0
        self.iStatus=0
        self.iChg=1
    def clrSchedule(self):
        """
        ### parameter
        ### return
            actions scheduled
        """
        self.iSchedule=0
        self.iChg=1
    def finStatus(self):
        """finalize status, property iStatus set to -2,
        to ensure final notification to be posted.
        IsActive set iStatus to -1.
        """
        self.iStatus=-2
        self.iChg=1
    def clrStatus(self):
        """clear status, notification turned off.
        """
        self.iStatus=-1
        self.iChg=1
    def IsChg(self):
        if self.iChg>0:
            self.iChg=0
            return 1
        return 0
    def IsActive(self,iHndFin=1):
        """check notification still active.
        event handler is supposed to use methode to prevent
        unnecessary notifications.
        """
        if self.iStatus>=0:
            return 1
        else:
            if self.iStatus==-2:
                if iHndFin>0:
                    self.clrStatus()
                return 1
            return 0
    def IncSchedule(self):
        """increment schedule counter.
        """
        self.iSchedule+=1
        self.iStatus=0
        self.iChg=1
    def IncStatus(self):
        """increment status counter.
        """
        self.iStatus+=1
        self.iChg=1
    def GetNtyDat(self):
        """get notify data.
        """
        if self.IsChg():
            oNtyDat=ldmGuiNtyDat(oNty=self)
            return oNtyDat
        return None
    def GetStatusOfs(self):
        """get string with schedule and status counter formated.
        schedule 2 digits,
        status 6 digits, or 'fin   ' in case iStatus==-2
        counter overflow is handled here.
        """
        if self.iSchedule>99:
            self.iSchedule=0
        if self.iStatus>999990:
            self.iStatus=0
        elif self.iStatus==-1:
            return '%02d.done   '%(self.iSchedule)
        elif self.iStatus==-2:
            return '%02d.fin    '%(self.iSchedule)
        return '%02d.%06d'%(self.iSchedule,self.iStatus)
    def SetPhase(self,sPhase):
        """set phase
        ### parameter
            sPhase      ... phase : string
        """
        if sPhase!=self.sPhase:
            self.iChg=1
        self.sPhase=sPhase
    def GetPhase(self):
        """get phase
        ### return
            sPhase      ... phase : string
        """
        return self.sPhase
    def SetStatus(self,sStatus):
        """set sStatus
        ### parameter
            sStatus     ... status : string
        """
        if self.sStatus!=sStatus:
            self.iChg=1
        self.sStatus=sStatus
    def GetStatus(self):
        """get sStatus
        ### return
            sStatus     ... status : string
        """
        return self.sStatus
    def SetVal(self,iVal,iMin=None,iMax=None):
        """set value
        ### parameter
            iVal    ... value : int
            iMin    ... minimum : int
            iMax    ... maximum : int
        """
        if iMin is not None:
            self.iMin=iMin
        if iMax is not None:
            self.iMax=iMax
        if iVal>self.iMax:
            iVal=self.iMax
        if iVal<self.iMin:
            iVal=self.iMin
        if self.iVal!=iVal:
            iChg=1
        self.iVal=iVal
    def SetMin(self,iMin):
        """set minimum
        ### parameter
            iMin    ... minimum : int
        """
        if self.iMin!=Min:
            self.iChg=1
        self.iMin=iMin
    def SetMax(self,iMax):
        """set maximum
        ### parameter
            iMax    ... maximum : int
        """
        if self.iMax!=iMax:
            self.iChg=1
        self.iMax=iMax
    def GetNormalized(self,rScale=1000):
        """get normalized value scaled between 0 to rScale.
        ### parameter
            rScale  ... scale limit
        ### return
            rVal    ... normalized value (min/max),
                        0 <= rVal <= rScale
        """
        try:
            if self.iVal<self.iMin:
                self.iVal=self.iMin
            if self.iVal>self.iMax:
                self.iVal=self.iMax
            rVal=(self.iVal-self.iMin)/(self.iMax-self.iMin)
            return rVal*rScale
        except:
            return 12

class ldmGuiNtyDat:
    def __init__(self,oNty=None):
        if oNty is None:
            self.sPhase=''
            self.sStatus=''
            self.sStatusOfs=''
            self.iValNormalized=0
        else:
            self.sPhase=oNty.GetPhase()[:]
            self.sStatus=oNty.GetStatus()[:]
            self.sStatusOfs=oNty.GetStatusOfs()[:]
            self.iValNormalized=oNty.GetNormalized()
    def GetPhase(self):
        """get phase
        ### return
            sPhase      ... phase : string
        """
        return self.sPhase
    def GetStatus(self):
        """get sStatus
        ### return
            sStatus     ... status : string
        """
        return self.sStatus
    def GetStatusOfs(self):
        """get string with schedule and status counter formated.
        schedule 2 digits,
        status 6 digits, or 'fin   ' in case iStatus==-2
        counter overflow is handled here.
        """
        return self.sStatusOfs
    def GetNormalized(self):
        """get normalized value scaled between 0 to 1000.
        ### return
            rVal    ... normalized value (min/max),
                        0 <= rVal <= 1000
        """
        return self.iValNormalized
